#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.autosign: Autosign tests for the cmdtest.py test suite
"""

import sys, os, time, shutil, atexit
from subprocess import run, DEVNULL
from pathlib import Path

from mmgen.cfg import Config
from mmgen.color import red, blue, yellow, cyan, orange, purple, gray
from mmgen.util import msg, suf, die, indent, fmt
from mmgen.led import LEDControl
from mmgen.autosign import Autosign, Signable

from ..include.common import (
	omsg,
	omsg_r,
	oqmsg,
	oqmsg_r,
	start_test_daemons,
	stop_test_daemons,
	joinpath,
	imsg,
	read_from_file,
	silence,
	end_silence,
	VirtBlockDevice,
)
from .include.common import ref_dir, dfl_words_file, dfl_bip39_file

from .base import CmdTestBase
from .include.input import stealth_mnemonic_entry

class CmdTestAutosignBase(CmdTestBase):
	networks     = ('btc',)
	tmpdir_nums  = [18]
	color        = True
	platform_skip = ('win32',)
	threaded     = False
	daemon_coins = []

	def __init__(self, cfg, trunner, cfgs, spawn):

		CmdTestBase.__init__(self, cfg, trunner, cfgs, spawn)

		if trunner is None:
			return

		self.silent_mount = self.live or self.tr.quiet
		self.network_ids = [c+'_tn' for c in self.daemon_coins] + self.daemon_coins

		self._create_autosign_instances(create_dirs=not cfg.skipping_deps)
		self.fs_image_path = Path(self.tmpdir).absolute() / 'removable_device_image'

		if sys.platform == 'linux':
			self.txdev = VirtBlockDevice(self.fs_image_path, '10M')

		if not (cfg.skipping_deps or self.live):
			self._create_removable_device()

		if sys.platform == 'darwin' and not cfg.no_daemon_stop:
			atexit.register(self._macOS_eject_disk, self.asi.dev_label)

		self.opts = ['--coins='+','.join(self.coins)]
		self.txhex_file = f'{self.tmpdir}/tx_dump.hex'

		if not self.live:
			self.spawn_env['MMGEN_TEST_SUITE_ROOT_PFX'] = self.tmpdir

		if self.threaded:
			self.spawn_env['MMGEN_TEST_SUITE_AUTOSIGN_THREADED'] = '1'

	def __del__(self):
		if hasattr(self, 'have_dummy_control_files'):
			db = LEDControl.boards['dummy']
			for fn in (db.control, db.trigger):
				run(f'sudo rm -f {fn}'.split(), check=True)

		if hasattr(self, 'txdev'):
			del self.txdev

		if not self.cfg.no_daemon_stop:
			if sys.platform == 'darwin':
				for label in (self.asi.dev_label, self.asi.ramdisk.label):
					self._macOS_eject_disk(label)

	def _create_autosign_instances(self, create_dirs):
		d = {'offline': {'name':'asi'}}

		if self.have_online:
			d['online'] =  {'name':'asi_online'}

		for subdir, data in d.items():
			asi = Autosign(
				Config({
					'coins': ','.join(self.coins),
					'test_suite': True,
					'test_suite_xmr_autosign': self.name == 'CmdTestXMRAutosign',
					'test_suite_autosign_threaded': self.threaded,
					'test_suite_root_pfx': None if self.live else self.tmpdir,
					'online': subdir == 'online',
				}))

			if create_dirs and not self.live:
				for k in ('mountpoint', 'shm_dir', 'wallet_dir'):
					if subdir == 'online' and k in ('shm_dir', 'wallet_dir'):
						continue
					if sys.platform == 'darwin' and k != 'mountpoint':
						continue
					getattr(asi, k).mkdir(parents=True, exist_ok=True)
					if sys.platform == 'darwin' and k == 'mountpoint':
						asi.mountpoint.rmdir()
						continue

			setattr(self, data['name'], asi)

	def _set_e2label(self, label):
		imsg(f'Setting label to {label}')
		for cmd in ('/sbin/e2label', 'e2label'):
			try:
				run([cmd, str(self.txdev.img_path), label], check=True)
				break
			except:
				if cmd == 'e2label':
					raise

	def _create_removable_device(self):
		if sys.platform == 'linux':
			self.txdev.create()
			self.txdev.attach(silent=True)
			args = [
				'-E', 'root_owner={}:{}'.format(os.getuid(), os.getgid()),
				'-L', self.asi.dev_label,
				str(self.txdev.img_path)]
			redir = DEVNULL
			for cmd in ('/sbin/mkfs.ext2', 'mkfs.ext2'):
				try:
					run([cmd] + args, stdout=redir, stderr=redir, check=True)
					break
				except:
					if cmd == 'mkfs.ext2':
						raise
			self.txdev.detach(silent=True)
		elif sys.platform == 'darwin':
			cmd = [
				'hdiutil', 'create', '-size', '10M', '-fs', 'exFAT',
				'-volname', self.asi.dev_label,
				str(self.fs_image_path)]
			redir = DEVNULL if self.tr.quiet else None
			run(cmd, stdout=redir, check=True)

	def _macOS_mount_fs_image(self, loc):
		time.sleep(0.2)
		run(
			['hdiutil', 'attach', '-mountpoint', str(loc.mountpoint), f'{self.fs_image_path}.dmg'],
			stdout=DEVNULL, check=True)

	def _macOS_eject_disk(self, label):
		try:
			run(['diskutil' , 'eject', label], stdout=DEVNULL, stderr=DEVNULL)
		except:
			pass

	def start_daemons(self):
		self.spawn(msg_only=True)
		start_test_daemons(*self.network_ids)
		return 'ok'

	def stop_daemons(self):
		self.spawn(msg_only=True)
		stop_test_daemons(*self.network_ids, remove_datadir=True)
		return 'ok'

	def run_setup(
			self,
			mn_type        = None,
			mn_file        = None,
			use_dfl_wallet = False,
			seed_len       = None,
			usr_entry_modes = False,
			wallet_passwd  = None,
			keylist_passwd = None,
			add_opts       = [],
			expect_args    = []):

		mn_desc = mn_type or 'default'
		mn_type = mn_type or 'mmgen'

		if sys.platform == 'darwin' and not self.cfg.no_daemon_stop:
			self._macOS_eject_disk(self.asi.ramdisk.label)

		self.insert_device()

		t = self.spawn(
			'mmgen-autosign',
			self.opts + add_opts
			+ ([] if mn_desc == 'default' else [f'--mnemonic-fmt={mn_type}'])
			+ ([f'--seed-len={seed_len}'] if seed_len else [])
			+ ['setup'],
			no_passthru_opts = True)

		if use_dfl_wallet:
			t.expect('Use default wallet for autosigning? (Y/n): ', 'y')
			t.passphrase('MMGen wallet', wallet_passwd)
		else:
			if use_dfl_wallet is not None: # None => no dfl wallet present
				t.expect('Use default wallet for autosigning? (Y/n): ', 'n')
			mn_file = mn_file or {'mmgen': dfl_words_file, 'bip39': dfl_bip39_file}[mn_type]
			mn = read_from_file(mn_file).strip().split()
			if not seed_len:
				t.expect('words: ', {12:'1', 18:'2', 24:'3'}[len(mn)])
				t.expect('OK? (Y/n): ', '\n')
			from mmgen.mn_entry import mn_entry
			entry_mode = 'full'
			mne = mn_entry(self.cfg, mn_type, entry_mode=entry_mode)
			if usr_entry_modes:
				t.expect('user-configured')
			else:
				t.expect(
					'Type a number.*: ',
					str(mne.entry_modes.index(entry_mode) + 1),
					regex = True)
			stealth_mnemonic_entry(t, mne, mn, entry_mode)

		t.written_to_file('Autosign wallet')

		if expect_args:
			t.expect(*expect_args)

		if keylist_passwd:
			t.passphrase('keylist data', keylist_passwd)
			t.expect('(y/N): ', 'y')

		t.read()
		self.remove_device()

		if sys.platform == 'darwin' and not self.cfg.no_daemon_stop:
			atexit.register(self._macOS_eject_disk, self.asi.ramdisk.label)

		return t

	def insert_device(self, asi='asi'):
		if self.live:
			return
		loc = getattr(self, asi)
		if sys.platform == 'linux':
			self._set_e2label(loc.dev_label)
			self.txdev.attach()
			for _ in range(20):
				if loc.device_inserted:
					break
				time.sleep(0.1)
			else:
				die(2, f'device insert timeout exceeded {loc.dev_label}')
		elif sys.platform == 'darwin':
			self._macOS_mount_fs_image(loc)

	def remove_device(self, asi='asi'):
		if self.live:
			return
		loc = getattr(self, asi)
		if sys.platform == 'linux':
			self.txdev.detach()
			for _ in range(20):
				if not loc.device_inserted:
					break
				time.sleep(0.1)
			else:
				die(2, f'device remove timeout exceeded {loc.dev_label}')
		elif sys.platform == 'darwin':
			self._macOS_eject_disk(loc.dev_label)

	def _mount_ops(self, loc, cmd, *args, **kwargs):
		return getattr(getattr(self, loc), cmd)(*args, silent=self.silent_mount, **kwargs)

	def do_mount(self, *args, **kwargs):
		return self._mount_ops('asi', 'do_mount', *args, **kwargs)

	def do_umount(self, *args, **kwargs):
		return self._mount_ops('asi', 'do_umount', *args, **kwargs)

	def _gen_listing(self):
		for k in self.asi.dirs:
			d = getattr(self.asi, k)
			if d.is_dir():
				yield '{:12} {}'.format(
					str(Path(*d.parts[6:])) + ':',
					' '.join(sorted(i.name for i in d.iterdir()))).strip()

class CmdTestAutosignClean(CmdTestAutosignBase):
	'autosign directory cleaning operations'

	have_online     = False
	live            = False
	simulate_led    = True
	no_insert_check = False
	coins           = ['btc']

	tmpdir_nums = [38]

	cmd_group = (
		('clean_no_xmr',   'cleaning signable file directories (no XMR)'),
		('clean_xmr_only', 'cleaning signable file directories (XMR-only)'),
		('clean_all',      'cleaning signable file directories (with XMR)'),
	)

	def create_fake_tx_files(self):
		imsg('Creating fake transaction files')

		if not self.asi.xmr_only:
			for fn in (
				'a.rawtx', 'a.sigtx',
				'b.rawtx', 'b.sigtx',
				'c.rawtx',
				'd.sigtx',
			):
				(self.asi.tx_dir / fn).touch()

			for fn in (
				'a.arawtx', 'a.asigtx', 'a.asubtx',
				'b.arawtx', 'b.asigtx',
				'c.asubtx',
				'd.arawtx', 'd.asubtx',
				'e.arawtx',
				'f.asigtx', 'f.asubtx',
			):
				(self.asi.txauto_dir / fn).touch()

			for fn in (
				'a.rawmsg.json', 'a.sigmsg.json',
				'b.rawmsg.json',
				'c.sigmsg.json',
				'd.rawmsg.json', 'd.sigmsg.json',
			):
				(self.asi.msg_dir / fn).touch()

		if self.asi.have_xmr:
			for fn in (
				'a.rawtx', 'a.sigtx', 'a.subtx',
				'b.rawtx', 'b.sigtx',
				'c.subtx',
				'd.rawtx', 'd.subtx',
				'e.rawtx',
				'f.sigtx', 'f.subtx',
			):
				(self.asi.xmr_tx_dir / fn).touch()

			for fn in (
				'a.raw', 'a.sig',
				'b.raw',
				'c.sig',
			):
				(self.asi.xmr_outputs_dir / fn).touch()

		return 'ok'

	def clean_no_xmr(self):
		return self._clean('btc,ltc,eth')

	def clean_xmr_only(self):
		self.asi = Autosign(Config({'_clone': self.asi.cfg, 'coins': 'xmr'}))
		return self._clean('xmr')

	def clean_all(self):
		self.asi = Autosign(Config({'_clone': self.asi.cfg, 'coins': 'xmr,btc,bch,eth'}))
		return self._clean('xmr,btc,bch,eth')

	def _clean(self, coins):

		self.spawn(msg_only=True)

		self.insert_device()

		silence()
		self.do_mount()
		end_silence()

		self.create_fake_tx_files()
		before = '\n'.join(self._gen_listing())

		t = self.spawn('mmgen-autosign', [f'--coins={coins}', 'clean'], no_msg=True)
		out = t.read()

		if sys.platform == 'darwin':
			self.insert_device()

		silence()
		self.do_mount()
		end_silence()

		after = '\n'.join(self._gen_listing())

		chk_non_xmr = """
			tx:          a.sigtx b.sigtx c.rawtx d.sigtx
			txauto:      a.asubtx b.asigtx c.asubtx d.asubtx e.arawtx f.asubtx
			msg:         a.sigmsg.json b.rawmsg.json c.sigmsg.json d.sigmsg.json
		"""
		chk_xmr = """
			xmr:         outputs tx
			xmr/tx:      a.subtx b.sigtx c.subtx d.subtx e.rawtx f.subtx
			xmr/outputs:
		"""
		chk = ''
		shred_count = 0

		if not self.asi.xmr_only:
			for k in ('tx_dir', 'txauto_dir', 'msg_dir'):
				shutil.rmtree(getattr(self.asi, k))
			chk += chk_non_xmr.rstrip()
			shred_count += 9

		if self.asi.have_xmr:
			shutil.rmtree(self.asi.xmr_dir)
			chk += chk_xmr.rstrip()
			shred_count += 9

		self.do_umount()
		self.remove_device()

		imsg(f'\nBefore cleaning:\n{before}')
		imsg(f'\nAfter cleaning:\n{after}')

		assert f'{shred_count} files shredded' in out
		assert after + '\n' == fmt(chk), f'\n{after}\n!=\n{fmt(chk)}'

		return t

class CmdTestAutosignThreaded(CmdTestAutosignBase):
	have_online     = True
	live            = False
	no_insert_check = False
	threaded        = True

	def _user_txcreate(
			self,
			user,
			progname    = 'txcreate',
			ui_handler  = None,
			chg_addr    = None,
			opts        = [],
			output_args = [],
			exit_val    = 0,
			inputs      = '1',
			tweaks      = [],
			expect_str  = None,
			data_arg    = None,
			need_rbf    = False):

		if output_args:
			assert not chg_addr

		if chg_addr:
			assert not output_args

		if need_rbf and not self.proto.cap('rbf'):
			return 'skip'

		def do_return():
			if expect_str:
				t.expect(expect_str)
			t.read()
			self.remove_device_online()
			return t

		self.insert_device_online()

		sid = self._user_sid(user)
		t = self.spawn(
			f'mmgen-{progname}',
			opts
			+ [f'--{user}', '--autosign']
			+ ([data_arg] if data_arg else [])
			+ (output_args or [f'{self.burn_addr},1.23456', f'{sid}:{chg_addr}']),
			exit_val = exit_val or None)

		if exit_val:
			return do_return()

		t = (ui_handler or self.txcreate_ui_common)(
			t,
			inputs          = inputs,
			interactive_fee = '32s',
			tweaks          = tweaks,
			file_desc       = 'Unsigned automount transaction')

		return do_return()

	def _user_txsend(
			self,
			user,
			*,
			comment   = None,
			no_wait   = False,
			need_rbf  = False,
			dump_hex  = False,
			mark_sent = False):

		if need_rbf and not self.proto.cap('rbf'):
			return 'skip'

		if not no_wait:
			self._wait_signed('transaction')

		extra_opt = (
			[f'--dump-hex={self.txhex_file}'] if dump_hex
			else ['--mark-sent'] if mark_sent
			else [])

		self.insert_device_online()
		t = self.spawn('mmgen-txsend',
			[f'--{user}', '--quiet', '--autosign'] + extra_opt, no_passthru_opts=['coin'])

		if mark_sent:
			t.written_to_file('Sent automount transaction')
		else:
			t.view_tx('t')
			t.do_comment(comment)
			if dump_hex:
				t.written_to_file('Serialized transaction hex data')
			else:
				self._do_confirm_send(t, quiet=True)
				t.written_to_file('Sent automount transaction')

		t.read()
		self.remove_device_online()
		return t

	def _wait_loop_start(self, add_opts=[]):
		t = self.spawn(
			'mmgen-autosign',
			self.opts + add_opts + ['--full-summary', 'wait'],
			direct_exec      = True,
			no_passthru_opts = True,
			spawn_env_override = self.spawn_env | {'EXEC_WRAPPER_DO_RUNTIME_MSG': ''})
		self.write_to_tmpfile('autosign_thread_pid', str(t.ep.pid))
		return t

	def wait_loop_start(self, add_opts=[]):
		import threading
		threading.Thread(
			target = self._wait_loop_start,
			kwargs = {'add_opts': add_opts},
			name   = 'Autosign wait loop').start()
		time.sleep(0.1) # try to ensure test output is displayed before next test starts
		return 'silent'

	def wait_loop_kill(self):
		return self._kill_process_from_pid_file('autosign_thread_pid', 'autosign wait loop')

	def _wait_signed(self, desc):
		oqmsg_r(gray(f'→ offline wallet{"s" if desc.endswith("s") else ""} waiting for {desc}'))
		assert not self.asi.device_inserted, f'‘{self.asi.dev_label}’ is inserted!'
		assert not self.asi.mountpoint.is_mount(), f'‘{self.asi.mountpoint}’ is mounted!'
		self.insert_device()
		while True:
			oqmsg_r(gray('.'))
			if self.asi.mountpoint.is_mount():
				oqmsg_r(gray(' signing '))
				break
			time.sleep(0.1)
		while True:
			oqmsg_r(gray('>'))
			if not self.asi.mountpoint.is_mount():
				oqmsg(gray(' done'))
				break
			time.sleep(0.1)
		imsg('')
		self.remove_device()
		return 'ok'

	def insert_device_online(self):
		return self.insert_device(asi='asi_online')

	def remove_device_online(self):
		return self.remove_device(asi='asi_online')

	def do_mount_online(self, *args, **kwargs):
		return self._mount_ops('asi_online', 'do_mount', *args, **kwargs)

	def do_umount_online(self, *args, **kwargs):
		return self._mount_ops('asi_online', 'do_umount', *args, **kwargs)

	async def txview(self):
		self.spawn(msg_only=True)
		self.insert_device()
		self.do_mount()
		src = Path(self.asi.txauto_dir)
		from mmgen.tx import CompletedTX
		txs = sorted(
			[await CompletedTX(cfg=self.cfg, filename=path, quiet_open=True) for path in sorted(src.iterdir())],
			key = lambda x: x.timestamp)
		for tx in txs:
			imsg(blue(f'\nViewing ‘{tx.infile.name}’:'))
			out = tx.info.format(terse=True)
			imsg(indent(out, indent='  '))
		self.do_umount()
		self.remove_device()
		return 'ok'

class CmdTestAutosign(CmdTestAutosignBase):
	'autosigning transactions for all supported coins'
	coins           = ['btc', 'bch', 'ltc', 'eth']
	daemon_coins    = ['btc', 'bch', 'ltc']
	txfile_coins    = ['btc', 'bch', 'ltc', 'eth', 'mm1', 'etc']
	have_online     = False
	live            = False
	simulate_led    = True
	no_insert_check = True
	wallet_passwd   = 'abc'

	filedir_map = (
		('btc', ''),
		('bch', ''),
		('ltc', 'litecoin'),
		('eth', 'ethereum'),
		('mm1', 'ethereum'),
		('etc', 'ethereum_classic'),
	)

	cmd_group = (
		('start_daemons',             'starting daemons'),
		('copy_tx_files',             'copying transaction files'),
		('gen_key',                   'generating key'),
		('create_dfl_wallet',         'creating default MMGen wallet'),
		('bad_opt1',                  'running ‘mmgen-autosign’ with --seed-len in invalid context'),
		('bad_opt2',                  'running ‘mmgen-autosign’ with --mnemonic-fmt in invalid context'),
		('bad_opt3',                  'running ‘mmgen-autosign’ with --led in invalid context'),
		('run_setup_dfl_wallet',      'running ‘autosign setup’ (with default wallet)'),
		('sign_quiet',                'signing transactions (--quiet)'),
		('remove_signed_txfiles',     'removing signed transaction files'),
		('run_setup_bip39',           'running ‘autosign setup’ (BIP39 mnemonic)'),
		('create_bad_txfiles',        'creating bad transaction files'),
		('sign_full_summary',         'signing transactions (--full-summary)'),
		('remove_signed_txfiles_btc', 'removing transaction files (BTC only)'),
		('remove_bad_txfiles',        'removing bad transaction files'),
		('sign_led',                  'signing transactions (--led - BTC files only)'),
		('remove_signed_txfiles',     'removing signed transaction files'),
		('sign_stealth_led',          'signing transactions (--stealth-led)'),
		('remove_signed_txfiles',     'removing signed transaction files'),
		('copy_msgfiles',             'copying message files'),
		('sign_quiet_msg',            'signing transactions and messages (--quiet)'),
		('remove_signed_txfiles',     'removing signed transaction files'),
		('create_bad_txfiles2',       'creating bad transaction files'),
		('remove_signed_msgfiles',    'removing signed message files'),
		('create_invalid_msgfile',    'creating invalid message file'),
		('sign_full_summary_msg',     'signing transactions and messages (--full-summary)'),
		('remove_invalid_msgfile',    'removing invalid message file'),
		('remove_bad_txfiles2',       'removing bad transaction files'),
		('sign_no_unsigned',          'signing transactions and messages (nothing to sign)'),
		('sign_no_unsigned_xmr',      'signing transactions and messages (nothing to sign, with XMR)'),
		('sign_no_unsigned_xmronly',  'signing transactions and messages (nothing to sign, XMR-only)'),
		('stop_daemons',              'stopping daemons'),
		('sign_bad_no_daemon',        'signing transactions (error, no daemons running)'),
		('wipe_key',                  'wiping the wallet encryption key'),
	)

	def __init__(self, cfg, trunner, cfgs, spawn):

		super().__init__(cfg, trunner, cfgs, spawn)

		if trunner is None:
			return

		if self.live and not self.cfg.exact_output:
			die(1, red('autosign_live tests must be run with --exact-output enabled!'))

		if self.no_insert_check:
			self.opts.append('--no-insert-check')

		self.tx_file_ops('set_count') # initialize self.tx_count here so we can resume anywhere
		self.bad_tx_count = 0

		def gen_msg_fns():
			fmap = dict(self.filedir_map)
			for coin in self.coins:
				if coin == 'xmr':
					continue
				sdir = os.path.join('test', 'ref', fmap[coin])
				for fn in os.listdir(sdir):
					if fn.endswith(f'[{coin.upper()}].rawmsg.json'):
						yield os.path.join(sdir, fn)

		self.ref_msgfiles = tuple(gen_msg_fns())
		self.good_msg_count = 0
		self.bad_msg_count = 0

		if self.simulate_led:
			db = LEDControl.boards['dummy']
			for fn in (db.control, db.trigger):
				run(f'sudo rm -f {fn}'.split(), check=True)
			LEDControl.create_dummy_control_files()
			usrgrp = {'linux': 'root:root', 'darwin': 'root:wheel'}[sys.platform]
			for fn in (db.control, db.trigger): # trigger the auto-chmod feature
				run(f'sudo chmod 644 {fn}'.split(), check=True)
				run(f'sudo chown {usrgrp} {fn}'.split(), check=True)
			self.have_dummy_control_files = True
			self.spawn_env['MMGEN_TEST_SUITE_AUTOSIGN_LED_SIMULATE'] = '1'

	def gen_key(self):
		self.insert_device()
		t = self.spawn('mmgen-autosign', self.opts + ['gen_key'])
		t.expect_getend('Wrote key file ')
		t.read()
		self.remove_device()
		return t

	def create_dfl_wallet(self):
		t = self.spawn('mmgen-walletconv', [
				f'--outdir={self.cfg.data_dir}',
				'--usr-randchars=0', '--quiet', '--hash-preset=1', '--label=foo',
				'test/ref/98831F3A.hex'
			]
		)
		t.passphrase_new('new MMGen wallet', self.wallet_passwd)
		t.written_to_file('MMGen wallet')
		return t

	def _bad_opt(self, cmdline, expect):
		t = self.spawn('mmgen-autosign', ['--coins=btc'] + cmdline, exit_val=1)
		t.expect(expect)
		return t

	def bad_opt1(self):
		return self._bad_opt(['--seed-len=128'], 'is valid')

	def bad_opt2(self):
		return self._bad_opt(['--mnemonic-fmt=bip39', 'wait'], 'is valid')

	def bad_opt3(self):
		return self._bad_opt(['--led', 'gen_key'], 'is not valid')

	def run_setup_dfl_wallet(self):
		return self.run_setup(mn_type='default', use_dfl_wallet=True, wallet_passwd=self.wallet_passwd)

	def run_setup_bip39(self):
		from mmgen.cfgfile import mmgen_cfg_file
		fn = mmgen_cfg_file(self.cfg, 'usr').fn
		old_data = mmgen_cfg_file(self.cfg, 'usr').get_data(fn)
		new_data = [d.replace('bip39:fixed', 'bip39:full')[2:]
			if d.startswith('# mnemonic_entry_modes') else d for d in old_data]
		with open(fn, 'w') as fh:
			fh.write('\n'.join(new_data) + '\n')
		t = self.run_setup(
			mn_type  = 'bip39',
			seed_len = 256,
			usr_entry_modes = True)
		with open(fn, 'w') as fh:
			fh.write('\n'.join(old_data) + '\n')
		return t

	def copy_tx_files(self):
		self.spawn(msg_only=True)
		return self.tx_file_ops('copy')

	def remove_signed_txfiles(self):
		self.tx_file_ops('remove_signed')
		return 'skip'

	def remove_signed_txfiles_btc(self):
		self.tx_file_ops('remove_signed', txfile_coins=['btc'])
		return 'skip'

	def tx_file_ops(self, op, txfile_coins=[]):

		assert op in ('copy', 'set_count', 'remove_signed')

		from .ref import CmdTestRef
		def gen():
			d = CmdTestRef.sources['ref_tx_file']
			dirmap = [e for e in self.filedir_map if e[0] in (txfile_coins or self.txfile_coins)]
			for coin, coindir in dirmap:
				for network in (0, 1):
					fn = d[coin][network]
					if fn:
						yield (coindir, fn)

		data = list(gen()) + [('', '25EFA3[2.34].testnet.rawtx')] # TX with 2 non-MMGen outputs

		self.tx_count = len(data)
		if op == 'set_count':
			return

		self.insert_device()

		silence()
		self.do_mount(verbose=not self.tr.quiet)
		end_silence()

		for coindir, fn in data:
			src = joinpath(ref_dir, coindir, fn)
			if self.cfg.debug_utf8:
				ext = '.testnet.rawtx' if fn.endswith('.testnet.rawtx') else '.rawtx'
				fn = fn[:-len(ext)] + '-α' + ext
			target = joinpath(self.asi.tx_dir, fn)
			if not op == 'remove_signed':
				shutil.copyfile(src, target)
			try:
				os.unlink(target.replace('.rawtx', '.sigtx'))
			except:
				pass

		self.do_umount()
		self.remove_device()

		return 'ok'

	def create_bad_txfiles(self):
		return self.bad_txfiles('create')

	def remove_bad_txfiles(self):
		return self.bad_txfiles('remove')

	create_bad_txfiles2 = create_bad_txfiles
	remove_bad_txfiles2 = remove_bad_txfiles

	def bad_txfiles(self, op):
		self.insert_device()
		self.do_mount()
		# create or delete 2 bad tx files
		self.spawn(msg_only=True)
		fns = [joinpath(self.asi.tx_dir, f'bad{n}.rawtx') for n in (1, 2)]
		if op == 'create':
			for fn in fns:
				with open(fn, 'w') as fp:
					fp.write('bad tx data\n')
			self.bad_tx_count = 2
		elif op == 'remove':
			for fn in fns:
				try:
					os.unlink(fn)
				except:
					pass
			self.bad_tx_count = 0
		self.do_umount()
		self.remove_device()
		return 'ok'

	def copy_msgfiles(self):
		return self.msgfile_ops('copy')

	def remove_signed_msgfiles(self):
		return self.msgfile_ops('remove_signed')

	def create_invalid_msgfile(self):
		return self.msgfile_ops('create_invalid')

	def remove_invalid_msgfile(self):
		return self.msgfile_ops('remove_invalid')

	def msgfile_ops(self, op):
		self.spawn(msg_only=True)
		destdir = joinpath(self.asi.mountpoint, 'msg')
		self.insert_device()
		self.do_mount()
		os.makedirs(destdir, exist_ok=True)
		if op.endswith('_invalid'):
			fn = os.path.join(destdir, 'DEADBE[BTC].rawmsg.json')
			if op == 'create_invalid':
				with open(fn, 'w') as fp:
					fp.write('bad data\n')
				self.bad_msg_count += 1
			elif op == 'remove_invalid':
				os.unlink(fn)
				self.bad_msg_count -= 1
		else:
			for fn in self.ref_msgfiles:
				if op == 'copy':
					if os.path.basename(fn) == 'ED405C[BTC].rawmsg.json': # contains bad Seed ID
						self.bad_msg_count += 1
					else:
						self.good_msg_count += 1
					imsg(f'Copying: {fn} -> {destdir}')
					shutil.copy2(fn, destdir)
				elif op == 'remove_signed':
					os.unlink(os.path.join(destdir, os.path.basename(fn).replace('rawmsg', 'sigmsg')))
		self.do_umount()
		self.remove_device()
		return 'ok'

	def do_sign(self, args=[], have_msg=False, exc_exit_val=None, expect_str=None):

		tx_desc = Signable.transaction.desc
		self.insert_device()

		def do_return():
			if expect_str:
				t.expect(expect_str)
			t.read()
			self.remove_device()
			imsg('')
			return t

		t = self.spawn(
				'mmgen-autosign',
				self.opts + args,
				exit_val = exc_exit_val or (1 if self.bad_tx_count or (have_msg and self.bad_msg_count) else None))

		if exc_exit_val:
			return do_return()

		t.expect(
			f'{self.tx_count} {tx_desc}{suf(self.tx_count)} signed' if self.tx_count else
			f'No unsigned {tx_desc}s')

		if self.bad_tx_count:
			t.expect(f'{self.bad_tx_count} {tx_desc}{suf(self.bad_tx_count)} failed to sign')

		if have_msg:
			t.expect(
				f'{self.good_msg_count} message file{suf(self.good_msg_count)}{{0,1}} signed'
					if self.good_msg_count else
				'No unsigned message files', regex=True)

			if self.bad_msg_count:
				t.expect(
					f'{self.bad_msg_count} message file{suf(self.bad_msg_count)}{{0,1}} failed to sign',
					regex = True)

		return do_return()

	def sign_quiet(self):
		return self.do_sign(['--quiet'])

	def sign_full_summary(self):
		return self.do_sign(['--full-summary'])

	def sign_led(self):
		return self.do_sign(['--quiet', '--led'])

	def sign_stealth_led(self):
		return self.do_sign(['--quiet', '--stealth-led'])

	def sign_quiet_msg(self):
		return self.do_sign(['--quiet'], have_msg=True)

	def sign_full_summary_msg(self):
		return self.do_sign(['--full-summary'], have_msg=True)

	def sign_bad_no_daemon(self):
		return self.do_sign(exc_exit_val=2, expect_str='listening on the correct port')

	def sign_no_unsigned(self):
		return self._sign_no_unsigned(
			coins   = 'BTC',
			present = ['non_xmr_signables'],
			absent  = ['xmr_signables'])

	def sign_no_unsigned_xmr(self):
		if self.coins == ['btc']:
			return 'skip'
		return self._sign_no_unsigned(
			coins = 'XMR,BTC',
			present = ['xmr_signables', 'non_xmr_signables'])

	def sign_no_unsigned_xmronly(self):
		if self.coins == ['btc']:
			return 'skip'
		return self._sign_no_unsigned(
			coins   = 'XMR',
			present = ['xmr_signables'],
			absent  = ['non_xmr_signables'])

	def _sign_no_unsigned(self, coins, present=[], absent=[]):
		self.insert_device()
		t = self.spawn('mmgen-autosign', ['--quiet', '--no-insert-check', f'--coins={coins}'])
		res = t.read()
		self.remove_device()
		for signable_list in present:
			for signable_clsname in getattr(Signable, signable_list):
				desc = getattr(Signable, signable_clsname).desc
				assert f'No unsigned {desc}s' in res, f'‘No unsigned {desc}s’ missing in output'
		for signable_list in absent:
			for signable_clsname in getattr(Signable, signable_list):
				desc = getattr(Signable, signable_clsname).desc
				assert not f'No unsigned {desc}s' in res, f'‘No unsigned {desc}s’ should be absent in output'
		return t

	def wipe_key(self):
		self.insert_device()
		t = self.spawn('mmgen-autosign', ['--quiet', '--no-insert-check', 'wipe_key'])
		t.expect('Shredding')
		t.read()
		self.remove_device()
		return t

class CmdTestAutosignBTC(CmdTestAutosign):
	'autosigning BTC transactions'
	coins        = ['btc']
	daemon_coins = ['btc']
	txfile_coins = ['btc']

class CmdTestAutosignLive(CmdTestAutosignBTC):
	'live autosigning BTC transactions'
	live            = True
	simulate_led    = False
	no_insert_check = False

	cmd_group = (
		('start_daemons',         'starting daemons'),
		('copy_tx_files',         'copying transaction files'),
		('gen_key',               'generating key'),
		('run_setup_mmgen',       'running ‘autosign setup’ (MMGen native mnemonic)'),
		('sign_live',             'signing transactions'),
		('create_bad_txfiles',    'creating bad transaction files'),
		('sign_live_led',         'signing transactions (--led)'),
		('remove_bad_txfiles',    'removing bad transaction files'),
		('sign_live_stealth_led', 'signing transactions (--stealth-led)'),
		('stop_daemons',          'stopping daemons'),
	)

	def __init__(self, cfg, trunner, cfgs, spawn):

		super().__init__(cfg, trunner, cfgs, spawn)

		if trunner is None:
			return

		try:
			led = LEDControl(enabled=True, simulate=self.simulate_led)
		except Exception as e:
			msg(str(e))
			die(2, 'LEDControl initialization failed')

		self.color = led.board.color

	def run_setup_mmgen(self):
		return self.run_setup(mn_type='mmgen', use_dfl_wallet=None)

	def sign_live(self):
		return self.do_sign_live()

	def sign_live_led(self):
		return self.do_sign_live(['--led'], f'The {self.color} LED should start blinking slowly now')

	def sign_live_stealth_led(self):
		return self.do_sign_live(['--stealth-led'], 'You should see no LED activity now')

	def do_sign_live(self, led_opts=None, led_msg=None):

		def prompt_remove():
			omsg_r(orange('\nExtract removable device and then hit ENTER '))
			input()

		def prompt_insert_sign(t):
			omsg(orange(insert_msg))
			t.expect(f'{self.tx_count} non-automount transactions signed')
			if self.bad_tx_count:
				t.expect(f'{self.bad_tx_count} non-automount transactions failed to sign')
			t.expect('Waiting')

		if led_opts:
			opts_msg = '‘' + ' '.join(led_opts) + '’'
			info_msg = 'Running ‘mmgen-autosign wait’ with {}. {}'.format(opts_msg, led_msg)
			insert_msg = f'Insert removable device and watch for fast {self.color} LED activity during signing'
		else:
			opts_msg = 'no LED'
			info_msg = 'Running ‘mmgen-autosign wait’'
			insert_msg = 'Insert removable device '

		self.spawn(msg_only=True)

		self.do_umount()
		prompt_remove()
		omsg('\n' + cyan(indent(info_msg)))

		t = self.spawn(
			'mmgen-autosign',
			self.opts + (led_opts or []) + ['--quiet', '--no-summary', 'wait'],
			no_msg   = True,
			exit_val = 1)

		if self.tr.quiet:
			omsg('')

		prompt_insert_sign(t)

		self.do_mount() # race condition due to device insertion detection
		self.remove_signed_txfiles()
		self.do_umount()

		imsg(purple('\nKilling wait loop!'))
		t.kill(2) # 2 = SIGINT

		if self.simulate_led and led_opts:
			t.expect(f'Resetting {self.color} LED')
		return t

class CmdTestAutosignLiveSimulate(CmdTestAutosignLive):
	'live autosigning BTC transactions with simulated LED support'
	simulate_led = True



#
# Adapted from: https://github.com/ethereum/pyethereum/blob/master/ethereum/utils.py
#   only funcs, vars required by vendored rlp module retained
#

import struct, functools
from typing import Any, Callable, TypeVar

ALL_BYTES = tuple(struct.pack('B', i) for i in range(256))

T = TypeVar('T')

def apply_to_return_value(callback: Callable[..., T]) -> Callable[..., Callable[..., T]]:

	def outer(fn):
		# We would need to type annotate *args and **kwargs but doing so segfaults
		# the PyPy builds. We ignore instead.
		@functools.wraps(fn)
		def inner(*args, **kwargs) -> T:  # type: ignore
			return callback(fn(*args, **kwargs))

		return inner

	return outer

to_list = apply_to_return_value(list)
to_set = apply_to_return_value(set)
to_dict = apply_to_return_value(dict)
to_tuple = apply_to_return_value(tuple)
to_list = apply_to_return_value(list)

def is_bytes(value: Any) -> bool:
	return isinstance(value, (bytes,bytearray))

def int_to_big_endian(value: int) -> bytes:
	return value.to_bytes((value.bit_length() + 7) // 8 or 1, 'big')

def big_endian_to_int(value: bytes) -> int:
	return int.from_bytes(value, 'big')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tool.file: Address and transaction file routines for the 'mmgen-tool' utility
"""

from .common import tool_cmd_base, options_annot_str

class tool_cmd(tool_cmd_base):
	"utilities for viewing/checking MMGen address and transaction files"

	need_proto = True

	def __init__(self, cfg, *, cmdname=None, proto=None, mmtype=None):
		if cmdname == 'txview':
			self.need_amt = True
		super().__init__(cfg=cfg, cmdname=cmdname, proto=proto, mmtype=mmtype)

	def _file_chksum(self, mmgen_addrfile, obj):
		kwargs = {'skip_chksum_msg':True}
		if not obj.__name__ == 'PasswordList':
			kwargs.update({'key_address_validity_check':False})
		ret = obj(self.cfg, self.proto, infile=mmgen_addrfile, **kwargs)
		if self.cfg.verbose:
			from ..util import msg, capfirst
			if ret.al_id.mmtype.name == 'password':
				msg('Passwd fmt:  {}\nPasswd len:  {}\nID string:   {}'.format(
					capfirst(ret.pw_info[ret.pw_fmt].desc),
					ret.pw_len,
					ret.pw_id_str))
			else:
				msg(f'Base coin:   {ret.base_coin} {capfirst(ret.network)}')
				msg(f'MMType:      {capfirst(ret.al_id.mmtype.name)}')
			msg(f'List length: {len(ret.data)}')
		return ret.chksum

	def addrfile_chksum(self, mmgen_addrfile: str):
		"compute checksum for MMGen address file"
		from ..addrlist import AddrList
		return self._file_chksum(mmgen_addrfile, AddrList)

	def keyaddrfile_chksum(self, mmgen_keyaddrfile: str):
		"compute checksum for MMGen key-address file"
		from ..addrlist import KeyAddrList
		return self._file_chksum(mmgen_keyaddrfile, KeyAddrList)

	def viewkeyaddrfile_chksum(self, mmgen_viewkeyaddrfile: str):
		"compute checksum for MMGen key-address file"
		from ..addrlist import ViewKeyAddrList
		return self._file_chksum(mmgen_viewkeyaddrfile, ViewKeyAddrList)

	def passwdfile_chksum(self, mmgen_passwdfile: str):
		"compute checksum for MMGen password file"
		from ..passwdlist import PasswordList
		return self._file_chksum(mmgen_passwdfile, PasswordList)

	async def txview(
			self,
			varargs_call_sig = { # hack to allow for multiple filenames - must be second argument!
				'args': (
					'mmgen_tx_file(s)',
					'pager',
					'terse',
					'sort',
					'filesort'),
				'dfls': (False, False, 'addr', 'mtime'),
				'annots': {
					'mmgen_tx_file(s)': str,
					'pager': 'send output to pager',
					'terse': 'produce compact tabular output',
					'sort':  'sort order for transaction inputs and outputs ' + options_annot_str(['addr', 'raw']),
					'filesort': 'file sort order ' + options_annot_str(['mtime', 'ctime', 'atime']),
				}
			},
			*infiles,
			**kwargs):
		"display specified raw or signed MMGen transaction files in human-readable form"

		terse = bool(kwargs.get('terse'))
		tx_sort = kwargs.get('sort') or 'addr'
		file_sort = kwargs.get('filesort') or 'mtime'

		from ..filename import MMGenFileList
		from ..tx import completed, CompletedTX
		flist = MMGenFileList(infiles, base_class=completed.Completed, proto=self.proto)
		flist.sort_by_age(key=file_sort) # in-place sort

		async def process_file(f):
			return (await CompletedTX(
				cfg        = self.cfg,
				filename   = f.name,
				quiet_open = True)).info.format(terse=terse, sort=tx_sort)

		return ('—'*77+'\n').join([await process_file(f) for f in flist]).rstrip()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.ref_3seed: Saved and generated reference file tests for 128,
                          192 and 256-bit seeds for the cmdtest.py test suite
"""

import os

from mmgen.util import msg, capfirst
from mmgen.wallet import get_wallet_cls

from ..include.common import cmp_or_die, joinpath
from .include.common import (
	pwfile,
	ref_wallet_hash_preset,
	ref_wallet_brainpass,
	ref_wallet_incog_offset,
	ref_dir,
	ok_msg
)
from .base import CmdTestBase
from .shared import CmdTestShared
from .wallet import CmdTestWalletConv

class CmdTestRef3Seed(CmdTestBase, CmdTestShared):
	'saved wallet files for 128-, 192- and 256-bit seeds + generated filename checks'
	networks = ('btc',)
	mmtypes = (None,)
	tmpdir_nums = [6, 7, 8]
	addr_idx_list_in = '1010,500-501,31-33,1,33,500,1011'
	pass_idx_list_in = '1,4,9-11,1100'
	chk_data = {
		'lens': (128, 192, 256),
		'sids': ('FE3C6545', '1378FC64', '98831F3A'),
	}
	shared_deps = ['mmdat', pwfile]
	skip_cmds = (
		'ref_xmrseed_25_passwdgen_1',
		'ref_xmrseed_25_passwdgen_2',
	)
	cmd_group = (
		# reading saved reference wallets
		('ref_wallet_chk',   ([], 'saved reference wallet')),
		('ref_seed_chk',     ([], 'saved seed file')),
		('ref_hex_chk',      ([], 'saved mmhex file')),
		('ref_plainhex_chk', ([], 'saved hex file')),
		('ref_dieroll_chk',  ([], 'saved dieroll (b6d) file')),
		('ref_mn_chk',       ([], 'saved native MMGen mnemonic file')),
		('ref_bip39_chk',    ([], 'saved BIP39 mnemonic file')),
		('ref_hincog_chk',   ([], 'saved hidden incog reference wallet')),
		('ref_brain_chk',    ([], 'saved brainwallet')),                    # in shared

		# generating new reference ('abc' brainwallet) wallets for filename checks:
		('ref_walletgen_brain',         ([], 'generating new reference wallet + filename check (brain)')),
		('ref_walletconv_words',        ([], 'wallet filename (native mnemonic)')),
		('ref_walletconv_bip39',        ([], 'wallet filename (bip39)')),
		('ref_walletconv_seed',         ([], 'wallet filename (seed)')),
		('ref_walletconv_hexseed',      ([], 'wallet filename (hex seed)')),
		('ref_walletconv_plainhexseed', ([], 'wallet filename (plain hex seed)')),
		('ref_walletconv_dieroll',      ([], 'wallet filename (dieroll (b6d) seed)')),
		('ref_walletconv_incog',        ([], 'wallet filename (incog)')),
		('ref_walletconv_hexincog',     ([], 'wallet filename (hex incog)')),
	)

	def __init__(self, cfg, trunner, cfgs, spawn):
		for k, _ in self.cmd_group:
			for n in (1, 2, 3): # 128, 192, 256 bits
				setattr(self, f'{k}_{n}', getattr(self, k))
		if cfgs:
			for n in self.tmpdir_nums:
				cfgs[str(n)]['addr_idx_list'] = self.addr_idx_list_in
				cfgs[str(n)]['pass_idx_list'] = self.pass_idx_list_in
		CmdTestBase.__init__(self, cfg, trunner, cfgs, spawn)

	def ref_wallet_chk(self):
		wf = joinpath(ref_dir, CmdTestWalletConv.sources[str(self.seed_len)]['ref_wallet'])
		return self.walletchk(wf, sid=self.seed_id)

	def ref_ss_chk(self, ss_type):
		ss = get_wallet_cls(ss_type)
		return self.walletchk(
			wf   = joinpath(ref_dir, f'{self.seed_id}.{ss.ext}'),
			wcls = ss,
			sid  = self.seed_id)

	def ref_seed_chk(self):
		return self.ref_ss_chk('seed')

	def ref_hex_chk(self):
		return self.ref_ss_chk('mmhex')

	def ref_plainhex_chk(self):
		return self.ref_ss_chk('plainhex')

	def ref_dieroll_chk(self):
		return self.ref_ss_chk('dieroll')

	def ref_mn_chk(self):
		return self.ref_ss_chk('words')

	def ref_bip39_chk(self):
		return self.ref_ss_chk('bip39')

	def ref_hincog_chk(self, desc='hidden incognito data'):
		source = CmdTestWalletConv.sources[str(self.seed_len)]
		for wtype, edesc, of_arg in (
				('hic_wallet',     '',             []),
				('hic_wallet_old', '(old format)', ['-O'])):
			ic_arg = ['-H{},{}'.format(
				joinpath(ref_dir, source[wtype]),
				ref_wallet_incog_offset)
			]
			slarg = [f'-l{self.seed_len} ']
			hparg = ['-p1']
			if wtype == 'hic_wallet_old' and self.cfg.profile:
				msg('')
			t = self.spawn('mmgen-walletchk',
				slarg + hparg + of_arg + ic_arg,
				extra_desc=edesc)
			t.passphrase(desc, self.wpasswd+'\n')
			if wtype == 'hic_wallet_old':
				t.expect('Is the Seed ID correct? (Y/n): ', '\n')
			chk = t.expect_getend('Seed ID: ')
			cmp_or_die(self.seed_id, chk)
			ok_msg()
		t.skip_ok = True
		return t

	def ref_walletgen_brain(self):
		sl_arg = f'-l{self.seed_len}'
		hp_arg = f'-p{ref_wallet_hash_preset}'
		label = f'ref. wallet (pw {ref_wallet_brainpass!r}, seed len {self.seed_len}) α'
		bf = 'ref.mmbrain'
		args = ['-d', self.tmpdir, hp_arg, sl_arg, '-ibw', '-L', label]
		self.write_to_tmpfile(bf, ref_wallet_brainpass)
		self.write_to_tmpfile(pwfile, self.wpasswd)
		t = self.spawn('mmgen-walletconv', self.testnet_opt + args + [self.usr_rand_arg], no_passthru_opts=True)
		t.license()
		t.expect('Enter brainwallet: ', ref_wallet_brainpass+'\n')
		ocls = get_wallet_cls('mmgen')
		t.passphrase_new('new '+ocls.desc, self.wpasswd)
		t.usr_rand(self.usr_rand_chars)
		fn = os.path.split(t.written_to_file(capfirst(ocls.desc)))[-1]
		import re
		idx = int(self.test_name[-1]) - 1
		pat = r'{}-[0-9A-F]{{8}}\[{},1\]{}.mmdat'.format(
			self.chk_data['sids'][idx],
			self.chk_data['lens'][idx],
			'-α' if self.cfg.debug_utf8 else '')
		assert re.match(pat, fn), f'{pat} != {fn}'
		sid = os.path.basename(fn.split('-')[0])
		cmp_or_die(sid, self.seed_id, desc='Seed ID')
		return t

	def ref_walletconv(self, ofmt, extra_args=[], re_pat=None):
		wf = self.get_file_with_ext('mmdat')
		pf = joinpath(self.tmpdir, pwfile)
		t = self.spawn('mmgen-walletconv', extra_args+['-d', 'test/trash', '-o', ofmt, '-P'+pf, wf])
		wcls = get_wallet_cls(fmt_code=ofmt)
		fn = os.path.split(t.written_to_file(capfirst(wcls.desc)))[-1]
		idx = int(self.test_name[-1]) - 1
		sid = self.chk_data['sids'][idx]
		slen = self.chk_data['lens'][idx]
		if re_pat:
			import re
			pat = re_pat.format(sid, slen)
			assert re.match(pat, fn), f'{pat} != {fn}'
		else:
			cmp_or_die('{}[{}]{}.{}'.format(
				sid,
				slen,
				'-α' if self.cfg.debug_utf8 else '',
				wcls.ext),
				fn)
		return t

	def ref_walletconv_words(self):
		return self.ref_walletconv(ofmt='mn')
	def ref_walletconv_bip39(self):
		return self.ref_walletconv(ofmt='bip39')
	def ref_walletconv_seed(self):
		return self.ref_walletconv(ofmt='mmseed')
	def ref_walletconv_hexseed(self):
		return self.ref_walletconv(ofmt='mmhex')
	def ref_walletconv_plainhexseed(self):
		return self.ref_walletconv(ofmt='hex')
	def ref_walletconv_dieroll(self):
		return self.ref_walletconv(ofmt='dieroll')

	def ref_walletconv_incog(self, ofmt='incog', ext='mmincog'):
		args = ['-r0', '-p1']
		pat = r'{}-[0-9A-F]{{8}}-[0-9A-F]{{8}}\[{},1\]' + ('-α' if self.cfg.debug_utf8 else '') + '.' + ext
		return self.ref_walletconv(ofmt=ofmt, extra_args=args, re_pat=pat)

	def ref_walletconv_hexincog(self):
		return self.ref_walletconv_incog(ofmt='incog_hex', ext='mmincox')

class CmdTestRef3Addr(CmdTestRef3Seed):
	'generated reference address and key-address files for 128-, 192- and 256-bit seeds'
	networks = ('btc', 'btc_tn', 'ltc', 'ltc_tn', 'bch', 'bch_tn')
	passthru_opts = ('coin', 'testnet', 'cashaddr')
	tmpdir_nums = [26, 27, 28]
	shared_deps = ['mmdat', pwfile]

	chk_data = {
		'lens': (128, 192, 256),
		'sids': ('FE3C6545', '1378FC64', '98831F3A'),
		'refaddrgen_legacy_1': {
			'btc': ('B230 7526 638F 38CB', 'A9DC 5A13 12CB 1317'),
			'bch': ('026D AFE0 8C60 6CFF', 'B406 4937 D884 6E48'),
			'ltc': ('2B23 5E97 848A B961', 'AEC3 E774 0B21 0202'),
		},
		'refaddrgen_segwit_1': {
			'btc': ('9914 6D10 2307 F348', '83C8 A6B6 ADA8 25B2'),
			'ltc': ('CC09 A190 B7DF B7CD', '0425 7893 C6F1 ECA3'),
		},
		'refaddrgen_bech32_1': {
			'btc': ('C529 D686 31AA ACD4', '21D0 26AD 3A22 5465'),
			'ltc': ('3DFB CFCC E180 DC9D', '8C72 D5C2 07E0 5F7B'),
		},
		'refaddrgen_compressed_1': {
			'btc': ('95EB 8CC0 7B3B 7856', '16E6 6170 154D 2202'),
			'bch': ('C560 A343 CEAB 118E', '3F56 8DC5 0383 CD78'),
			'ltc': ('35D5 8ECA 9A42 46C3', '15B3 5492 D3D3 6854'),
		},
		'refkeyaddrgen_legacy_1': {
			'btc': ('CF83 32FB 8A8B 08E2', '1F67 B73A FF8C 5D15'),
			'bch': ('6909 4C64 119A 7681', '7E48 5071 5E41 D1AE'),
			'ltc': ('1896 A26C 7F14 2D01', 'FA0E CD4E ADAF DBF4'),
		},
		'refkeyaddrgen_compressed_1': {
			'btc': ('E43A FA46 5751 720A', 'FDEE 8E45 1C0A 02AD'),
			'bch': ('7068 9B37 8ABF 3E31', 'C688 29A5 BA4C 21B2'),
			'ltc': ('7603 2FE3 2145 FFAD', '3FE0 5A8E 5FBE FF3E'),
		},
		'refkeyaddrgen_segwit_1': {
			'btc': ('C13B F717 D4E8 CF59', 'BB71 175C 5416 19D8'),
			'ltc': ('054B 9794 55B4 5D82', 'DE85 3CF3 9636 FE2E'),
		},
		'refkeyaddrgen_bech32_1': {
			'btc': ('934F 1C33 6C06 B18C', 'A283 5BAB 7AF3 3EA4'),
			'ltc': ('A6AD DF53 5968 7B6A', '9572 43E0 A4DC 0B2E'),
		},
		'refaddrgen_legacy_2': {
			'btc': ('8C17 A5FA 0470 6E89', '764C 66F9 7502 AAEA'),
			'bch': ('8117 24B6 3FDA 6B40', 'E58C A8A4 C371 66AE'),
			'ltc': ('2B77 A009 D5D0 22AD', '51D1 979D 0A35 F24B'),
		},
		'refaddrgen_compressed_2': {
			'btc': ('2615 8401 2E98 7ECA', 'A386 EE07 A356 906D'),
			'bch': ('3364 0F9D 8355 2A53', '3451 F741 0A8A FA56'),
			'ltc': ('197C C48C 3C37 AB0F', '8DDC 5FE3 BFF9 1226'),
		},
		'refaddrgen_segwit_2': {
			'btc': ('91C4 0414 89E4 2089', 'BF9F C67F ED22 A47B'),
			'ltc': ('8F12 FA7B 9F12 594C', '2609 8494 A23C F836'),
		},
		'refaddrgen_bech32_2': {
			'btc': ('2AA3 78DF B965 82EB', '027B 1C1F 7FB2 D859'),
			'ltc': ('951C 8FB2 FCA5 87D1', '4A5D 67E0 8210 FEF2'),
		},
		'refkeyaddrgen_legacy_2': {
			'btc': ('9648 5132 B98E 3AD9', '1BD3 5A36 D51C 256D'),
			'bch': ('C4D8 7C36 DC77 F8C2', '953D 245C 8CFF AC72'),
			'ltc': ('DBD4 FAB6 7E46 CD07', '8822 3FDF FEC0 6A8C'),
		},
		'refkeyaddrgen_compressed_2': {
			'btc': ('6D6D 3D35 04FD B9C3', '94BF 4BCF 10B2 394B'),
			'bch': ('3E7F C369 2AB9 BD58', '0C99 14CD 5ADE 6782'),
			'ltc': ('F5DA 9D60 6798 C4E9', '7918 88DE 9096 DD7A'),
		},
		'refkeyaddrgen_segwit_2': {
			'btc': ('C98B DF08 A3D5 204B', '7E7F DF50 FE04 6F68'),
			'ltc': ('1829 7FE7 2567 CB91', 'BE92 D19C 7589 EF30'),
		},
		'refkeyaddrgen_bech32_2': {
			'btc': ('4A6B 3762 DF30 9368', '12DD 1888 36BA 85F7'),
			'ltc': ('5C12 FDD4 17AB F179', 'E195 B28C 59C4 C5EC'),
		},
		'refaddrgen_legacy_3': {
			'btc': ('6FEF 6FB9 7B13 5D91', '424E 4326 CFFE 5F51'),
			'bch': ('E580 43BB 0F96 AA93', '630E 174A 8DDE 1BCE'),
			'ltc': ('AD52 C3FE 8924 AAF0', '4EBE 2E85 E969 1B30'),
		},
		'refaddrgen_compressed_3': {
			'btc': ('A33C 4FDE F515 F5BC', '6C48 AA57 2056 C8C8'),
			'bch': ('E37B AF41 7997 A28C', '0D5D 9A58 D6E9 92EE'),
			'ltc': ('3FC0 8F03 C2D6 BD19', '4C0A 49B6 2DD1 1BE0'),
		},
		'refaddrgen_segwit_3': {
			'btc': ('06C1 9C87 F25C 4EE6', '072C 8B07 2730 CB7A'),
			'ltc': ('63DF E42A 0827 21C3', '5DD1 D186 DBE1 59F2'),
		},
		'refaddrgen_bech32_3': {
			'btc': ('9D2A D4B6 5117 F02E', '0527 9C39 6C1B E39A'),
			'ltc': ('FF1C 7939 5967 AB82', 'ED3D 8AA4 BED4 0B40'),
		},
		'refkeyaddrgen_legacy_3': {
			'btc': ('9F2D D781 1812 8BAD', '88CC 5120 9A91 22C2'),
			'bch': ('A0EE B039 48F4 24AE', 'B014 E0AB 5F87 EC64'),
			'ltc': ('B804 978A 8796 3ED4', '98B5 AC35 F334 0398'),
		},
		'refkeyaddrgen_compressed_3': {
			'btc': ('420A 8EB5 A9E2 7814', 'F43A CB4A 81F3 F735'),
			'bch': ('33E7 5C06 88CF 2792', '6E09 FF73 B7C8 00D4'),
			'ltc': ('8D1C 781F EB7F 44BC', '05F3 5C68 FD31 FCEF'),
		},
		'refkeyaddrgen_segwit_3': {
			'btc': ('A447 12C2 DD14 5A9B', 'C770 7391 C415 21F9'),
			'ltc': ('E8A3 9F6E E164 A521', 'D3D5 BFDD F5D5 20BD'),
		},
		'refkeyaddrgen_bech32_3': {
			'btc': ('D0DD BDE3 87BE 15AE', '7552 D70C AAB8 DEAA'),
			'ltc': ('74A0 7DD5 963B 6326', '2CDA A007 4B9F E9A5'),
		},
	}

	cmd_group = (
		('ref_walletgen_brain',       ([], 'generating new reference wallet + filename check (brain)')),
		('refaddrgen_legacy',         ([], 'new refwallet addr chksum (uncompressed)')),
		('refaddrgen_compressed',     ([], 'new refwallet addr chksum (compressed)')),
		('refaddrgen_segwit',         ([], 'new refwallet addr chksum (segwit)')),
		('refaddrgen_bech32',         ([], 'new refwallet addr chksum (bech32)')),
		('refkeyaddrgen_legacy',      ([], 'new refwallet key-addr chksum (uncompressed)')),
		('refkeyaddrgen_compressed',  ([], 'new refwallet key-addr chksum (compressed)')),
		('refkeyaddrgen_segwit',      ([], 'new refwallet key-addr chksum (segwit)')),
		('refkeyaddrgen_bech32',      ([], 'new refwallet key-addr chksum (bech32)')),
	)

	def call_addrgen(self, mmtype, name='addrgen'):
		wf = self.get_file_with_ext('mmdat')
		return getattr(self, name)(wf, check_ref=True, mmtype=mmtype)

	def refaddrgen_legacy(self):
		return self.call_addrgen('legacy')

	def refaddrgen_compressed(self):
		return self.call_addrgen('compressed')

	def refaddrgen_segwit(self):
		if self.proto.cap('segwit'):
			return self.call_addrgen('segwit')
		return 'skip'

	def refaddrgen_bech32(self):
		if self.proto.cap('segwit'):
			return self.call_addrgen('bech32')
		return 'skip'

	def refkeyaddrgen_legacy(self):
		return self.call_addrgen('legacy', 'keyaddrgen')

	def refkeyaddrgen_compressed(self):
		return self.call_addrgen('compressed', 'keyaddrgen')

	def refkeyaddrgen_segwit(self):
		if self.proto.cap('segwit'):
			return self.call_addrgen('segwit', 'keyaddrgen')
		return 'skip'

	def refkeyaddrgen_bech32(self):
		if self.proto.cap('segwit'):
			return self.call_addrgen('bech32', 'keyaddrgen')
		return 'skip'

class CmdTestRef3Passwd(CmdTestRef3Seed):
	'generated reference password files for 128-, 192- and 256-bit seeds'
	tmpdir_nums = [26, 27, 28]
	shared_deps = ['mmdat', pwfile]

	chk_data = {
		'lens': (128, 192, 256),
		'sids': ('FE3C6545', '1378FC64', '98831F3A'),
		'refpasswdgen_1':               'EB29 DC4F 924B 289F',
		'refpasswdgen_half_1':          'D310 2593 B5D9 2E88',
		'ref_b32passwdgen_1':           '37B6 C218 2ABC 7508',
		'ref_hexpasswdgen_1':           '8E99 E696 84CE E7D5',
		'ref_hexpasswdgen_half_1':      '8E99 E696 84CE E7D5',
		'ref_bip39_12_passwdgen_1':     '834F CF45 0B33 8AF0',
		'ref_bip39_18_passwdgen_1':     '834F CF45 0B33 8AF0',
		'ref_bip39_24_passwdgen_1':     '834F CF45 0B33 8AF0',
		'ref_hex2bip39_24_passwdgen_1': '91AF E735 A31D 72A0',
		'refpasswdgen_2':               'ADEA 0083 094D 489A',
		'refpasswdgen_half_2':          '12B3 4929 9506 76E0',
		'ref_b32passwdgen_2':           '2A28 C5C7 36EC 217A',
		'ref_hexpasswdgen_2':           '88F9 0D48 3A7E 7CC2',
		'ref_hexpasswdgen_half_2':      '59F3 8F48 861E 1186',
		'ref_bip39_12_passwdgen_2':     'D32D B8D7 A840 250B',
		'ref_bip39_18_passwdgen_2':     '0FAA 78DD A6BA 31AD',
		'ref_bip39_24_passwdgen_2':     '0FAA 78DD A6BA 31AD',
		'ref_hex2bip39_24_passwdgen_2': '0E8E 23C9 923F 7C2D',
		'refpasswdgen_3':               '2D6D 8FBA 422E 1315',
		'refpasswdgen_half_3':          '272C B770 0176 D7EA',
		'ref_b32passwdgen_3':           'F6C1 CDFB 97D9 FCAE',
		'ref_hexpasswdgen_3':           'BD4F A0AC 8628 4BE4',
		'ref_hexpasswdgen_half_3':      'FBDD F733 FFB9 21C1',
		'ref_bip39_12_passwdgen_3':     'A86E EA14 974A 1B0E',
		'ref_bip39_18_passwdgen_3':     'EF87 9904 88E2 5884',
		'ref_bip39_24_passwdgen_3':     'EBE8 2A8F 8F8C 7DBD',
		'ref_hex2bip39_24_passwdgen_3': '93FA 5EFD 33F3 760E',
		'ref_xmrseed_25_passwdgen_3':   '91AE E76A 2827 C8CC',
	}

	cmd_group = (
		('ref_walletgen_brain',        ([], 'generating new reference wallet + filename check (brain)')),
		('refpasswdgen',               ([], 'new refwallet passwd file chksum')),
		('refpasswdgen_half',          ([], 'new refwallet passwd file chksum (half-length)')),
		('ref_b32passwdgen',           ([], 'new refwallet passwd file chksum (base32)')),
		('ref_hexpasswdgen',           ([], 'new refwallet passwd file chksum (hex)')),
		('ref_hexpasswdgen_half',      ([], 'new refwallet passwd file chksum (hex, half-length)')),
		('ref_bip39_12_passwdgen',     ([], 'new refwallet passwd file chksum (BIP39, 12 words)')),
		('ref_bip39_18_passwdgen',     ([], 'new refwallet passwd file chksum (BIP39, up to 18 words)')),
		('ref_bip39_24_passwdgen',     ([], 'new refwallet passwd file chksum (BIP39, up to 24 words)')),
		('ref_xmrseed_25_passwdgen',   ([], 'new refwallet passwd file chksum (Monero 25-word mnemonic)')),
		('ref_hex2bip39_24_passwdgen', ([], 'new refwallet passwd file chksum (hex-to-BIP39, up to 24 words)')),
	)

	def pwgen(self, ftype, id_str, pwfmt=None, pwlen=None, extra_opts=[], stdout=False):
		wf = self.get_file_with_ext('mmdat')
		pwfmt = ([f'--passwd-fmt={pwfmt}'] if pwfmt else [])
		pwlen = ([f'--passwd-len={pwlen}'] if pwlen else [])
		return self.addrgen(
			wf,
			check_ref  = True,
			ftype      = ftype,
			id_str     = id_str,
			extra_opts = pwfmt + pwlen + extra_opts,
			stdout     = stdout,
			no_passthru_opts = True)

	def refpasswdgen(self):
		return self.pwgen('pass', 'alice@crypto.org')

	def refpasswdgen_half(self):
		return self.pwgen('pass', 'alice@crypto.org', pwlen='h')

	def ref_b32passwdgen(self):
		return self.pwgen('pass32', 'фубар@crypto.org', 'b32', 17)

	def ref_hexpasswdgen(self):
		pwlen = {'1':32, '2':48, '3':64}[self.test_name[-1]]
		return self.pwgen('passhex', 'фубар@crypto.org', 'hex', pwlen)

	def ref_hexpasswdgen_half(self):
		return self.pwgen('passhex', 'фубар@crypto.org', 'hex', 'h', ['--accept-defaults'], stdout=True)

	def mn_pwgen(self, pwlen, pwfmt, ftype='passbip39'):
		if pwlen > {'1':12, '2':18, '3':24}[self.test_name[-1]]:
			return 'skip'
		if pwfmt == 'xmrseed':
			if self.cfg.no_altcoin:
				return 'skip'
			pwlen += 1
		return self.pwgen(ftype, 'фубар@crypto.org', pwfmt, pwlen, ['--accept-defaults'])

	def ref_bip39_12_passwdgen(self):
		return self.mn_pwgen(12, 'bip39')

	def ref_bip39_18_passwdgen(self):
		return self.mn_pwgen(18, 'bip39')

	def ref_bip39_24_passwdgen(self):
		return self.mn_pwgen(24, 'bip39')

	def ref_hex2bip39_24_passwdgen(self):
		return self.mn_pwgen(24, 'hex2bip39')

	def ref_xmrseed_25_passwdgen(self):
		return self.mn_pwgen(24, 'xmrseed', ftype='passxmrseed')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
rpc.util: RPC library utility functions for the MMGen Project
"""

import re, json
from collections import namedtuple

from ..util import msg, pp_fmt
from ..objmethods import HiliteStr, InitErrors

def dmsg_rpc(fs, data=None, *, is_json=False):
	msg(
		fs if data is None else
		fs.format(pp_fmt(json.loads(data) if is_json else data))
	)

def dmsg_rpc_backend(host_url, host_path, payload):
	msg(
		f'\n    RPC URL: {host_url}{host_path}' +
		'\n    RPC PAYLOAD data (httplib) ==>' +
		f'\n{pp_fmt(payload)}\n')

def noop(*args, **kwargs):
	pass

auth_data = namedtuple('rpc_auth_data', ['user', 'passwd'])

class json_encoder(json.JSONEncoder):
	def default(self, o):
		if type(o).__name__.endswith('Amt'):
			return str(o)
		else:
			return json.JSONEncoder.default(self, o)

class IPPort(HiliteStr, InitErrors):
	color = 'yellow'
	width = 0
	trunc_ok = False
	min_len = 9  # 0.0.0.0:0
	max_len = 21 # 255.255.255.255:65535
	def __new__(cls, s):
		if isinstance(s, cls):
			return s
		try:
			m = re.fullmatch(r'{q}\.{q}\.{q}\.{q}:(\d{{1,10}})'.format(q=r'([0-9]{1,3})'), s)
			assert m is not None, f'{s!r}: invalid IP:HOST specifier'
			for e in m.groups():
				if len(e) != 1 and e[0] == '0':
					raise ValueError(f'{e}: leading zeroes not permitted in dotted decimal element or port number')
			res = [int(e) for e in m.groups()]
			for e in res[:4]:
				assert e <= 255, f'{e}: dotted decimal element > 255'
			assert res[4] <= 65535, f'{res[4]}: port number > 65535'
			me = str.__new__(cls, s)
			me.ip = '{}.{}.{}.{}'.format(*res)
			me.ip_num = sum(res[i] * (2 ** (-(i-3)*8)) for i in range(4))
			me.port = res[4]
			return me
		except Exception as e:
			return cls.init_fail(e, s)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.tw.view: Bitcoin base protocol base class for tracking wallet view classes
"""

class BitcoinTwView:

	txid_w = 64
	txid_max_w  = {'txid': 64}
	txid_min_w  = {'txid': 7}
	txid_nice_w = {'txid': 12}
	nice_addr_w = {'addr': 16}

	age_col_params = {
		'confs':     (7,  'Confs'),
		'block':     (8,  'Block'),
		'days':      (6,  'Age(d)'),
		'date':      (8,  'Date'),
		'date_time': (16, 'Date/Time'),
	}



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.btc.rpc.local: Bitcoin base protocol local RPC client for the MMGen Project
"""

import os

from ....base_obj import AsyncInit
from ....obj import TrackingWalletName
from ....util import ymsg, die, fmt
from ....fileutil import get_lines_from_file
from ....rpc.local import RPCClient
from ....rpc.util import auth_data

no_credentials_errmsg = """
	Error: no {proto_name} RPC authentication method found

	RPC credentials must be supplied using one of the following methods:

	1) If daemon is local and running as same user as you:

	   - no credentials required, or matching rpcuser/rpcpassword and
	     rpc_user/rpc_password values in {cf_name}.conf and mmgen.cfg

	2) If daemon is running remotely or as different user:

	   - matching credentials in {cf_name}.conf and mmgen.cfg as described
	     above

	The --rpc-user/--rpc-password options may be supplied on the MMGen command
	line.  They override the corresponding values in mmgen.cfg. Set them to an
	empty string to use cookie authentication with a local server when the
	options are set in mmgen.cfg.

	For better security, rpcauth should be used in {cf_name}.conf instead of
	rpcuser/rpcpassword.
"""

class CallSigs:

	class bitcoin_core:

		def __init__(self, cfg):
			self.cfg = cfg

		def createwallet(
				self,
				wallet_name,
				*,
				no_keys         = True,
				blank           = True,
				passphrase      = '',
				descriptors     = True,
				load_on_startup = True):
			"""
			Quirk: when --datadir is specified (even if standard), wallet is created directly in
			datadir, otherwise in datadir/wallets
			"""
			return (
				'createwallet',
				wallet_name,     # 1. wallet_name
				no_keys,         # 2. disable_private_keys
				blank,           # 3. blank (no keys or seed)
				passphrase,      # 4. passphrase (empty string for non-encrypted)
				False,           # 5. avoid_reuse (track address reuse)
				descriptors,     # 6. descriptors (native descriptor wallet)
				load_on_startup) # 7. load_on_startup

		def gettransaction(self, txid, include_watchonly, verbose):
			return (
				'gettransaction',
				txid,               # 1. transaction id
				include_watchonly,  # 2. optional, default=true for watch-only wallets, otherwise false
				verbose)            # 3. optional, default=false -- include a `decoded` field containing
									# => the decoded transaction (equivalent to RPC decoderawtransaction)

	class litecoin_core(bitcoin_core):

		def createwallet(
				self,
				wallet_name,
				*,
				no_keys         = True,
				blank           = True,
				passphrase      = '',
				descriptors     = True,
				load_on_startup = True):
			return (
				'createwallet',
				wallet_name,    # 1. wallet_name
				no_keys,        # 2. disable_private_keys
				blank)          # 3. blank (no keys or seed)

		def gettransaction(self, txid, include_watchonly, verbose):
			return (
				'gettransaction',
				txid,               # 1. transaction id
				include_watchonly)  # 2. optional, default=true for watch-only wallets, otherwise false

	class bitcoin_cash_node(litecoin_core):
		pass

class BitcoinRPCClient(RPCClient, metaclass=AsyncInit):

	auth_type = 'basic'
	has_auth_cookie = True
	wallet_path = '/'
	dfl_twname = 'mmgen-tracking-wallet'

	async def __init__(
			self,
			cfg,
			proto,
			*,
			daemon,
			backend,
			ignore_wallet):

		self.proto = proto
		self.daemon = daemon
		self.call_sigs = getattr(CallSigs, daemon.id)(cfg)
		self.twname = TrackingWalletName(cfg.regtest_user or proto.tw_name or cfg.tw_name or self.dfl_twname)

		super().__init__(
			cfg  = cfg,
			host = proto.rpc_host or cfg.rpc_host or 'localhost',
			port = daemon.rpc_port)

		self.set_auth()

		await self.set_backend_async(backend) # backend requires self.auth

		self.cached = {}

		self.caps = ('full_node',)
		for func, cap in (
			('setlabel', 'label_api'),
			('getdeploymentinfo', 'deployment_info'),
			('signrawtransactionwithkey', 'sign_with_key')):
			if len((await self.call('help', func)).split('\n')) > 3:
				self.caps += (cap,)

		call_group = [
			('getblockcount', ()),
			('getblockhash', (0,)),
			('getnetworkinfo', ()),
			('getblockchaininfo', ()),
		] + (
			[('getdeploymentinfo', ())] if 'deployment_info' in self.caps else []
		)

		(
			self.blockcount,
			block0,
			self.cached['networkinfo'],
			self.cached['blockchaininfo'],
			self.cached['deploymentinfo'],
		) = (
			await self.gathered_call(None, tuple(call_group))
		) + (
			[] if 'deployment_info' in self.caps else [None]
		)

		self.daemon_version = self.cached['networkinfo']['version']
		self.daemon_version_str = self.cached['networkinfo']['subversion']
		self.chain = self.cached['blockchaininfo']['chain']

		tip = await self.call('getblockhash', self.blockcount)
		self.cur_date = (await self.call('getblockheader', tip))['time']
		if self.chain != 'regtest':
			self.chain += 'net'
		assert self.chain in self.proto.networks

		async def check_chainfork_mismatch(block0):
			try:
				if block0 != self.proto.block0:
					raise ValueError(f'Invalid Genesis block for {self.proto.cls_name} protocol')
				for fork in self.proto.forks:
					if fork.height is None or self.blockcount < fork.height:
						break
					if fork.hash != await self.call('getblockhash', fork.height):
						die(3, f'Bad block hash at fork block {fork.height}. Is this the {fork.name} chain?')
			except Exception as e:
				die(2, '{!s}\n{c!r} requested, but this is not the {c} chain!'.format(e, c=self.proto.coin))

		if self.chain == 'mainnet': # skip this for testnet, as Genesis block may change
			await check_chainfork_mismatch(block0)

		if not ignore_wallet:
			await self.check_or_create_daemon_wallet()

		# for regtest, wallet_path must remain '/' until Carol’s user wallet has been created
		if self.chain != 'regtest' or cfg.regtest_user:
			self.wallet_path = f'/wallet/{self.twname}'

	@property
	async def walletinfo(self):
		if not hasattr(self, '_walletinfo'):
			self._walletinfo = await self.call('getwalletinfo')
		return self._walletinfo

	def set_auth(self):
		"""
		MMGen's credentials override coin daemon's
		"""
		if self.cfg.network == 'regtest':
			from ..regtest import MMGenRegtest
			user = MMGenRegtest.rpc_user
			passwd = MMGenRegtest.rpc_password
		else:
			user = (
				self.proto.rpc_user or self.cfg.rpc_user or self.get_daemon_cfg_option('rpcuser')
				or self.daemon.rpc_user)
			passwd = (
				self.proto.rpc_password or self.cfg.rpc_password or self.get_daemon_cfg_option('rpcpassword')
				or self.daemon.rpc_password)

		if user and passwd:
			self.auth = auth_data(user, passwd)
			return

		if self.has_auth_cookie:
			if cookie := self.get_daemon_auth_cookie():
				self.auth = auth_data(*cookie.split(':'))
				return

		die(1, '\n\n' + fmt(no_credentials_errmsg, strip_char='\t', indent='  ').format(
			proto_name = self.proto.name,
			cf_name = (self.proto.is_fork_of or self.proto.name).lower()))

	def make_host_path(self, wallet):
		return f'/wallet/{wallet}' if wallet else self.wallet_path

	@property
	async def tracking_wallet_exists(self):
		return self.twname in [i['name'] for i in (await self.call('listwalletdir'))['wallets']]

	async def check_or_create_daemon_wallet(self):

		if self.chain == 'regtest' and self.cfg.regtest_user != 'carol':
			return

		loaded_wnames = await self.call('listwallets')

		if self.twname not in loaded_wnames:
			wnames = [i['name'] for i in (await self.call('listwalletdir'))['wallets']]
			if self.twname in wnames:
				await self.call('loadwallet', self.twname)
			else:
				await self.icall('createwallet', wallet_name=self.twname)
				ymsg(f'Created {self.daemon.coind_name} wallet {self.twname!r}')

	def get_daemon_cfg_fn(self):
		# Use dirname() to remove 'bob' or 'alice' component
		return os.path.join(
			(os.path.dirname(self.cfg.data_dir) if self.proto.regtest else self.daemon.datadir),
			self.daemon.cfg_file)

	def get_daemon_cfg_option(self, req_key):
		return list(self.get_daemon_cfg_options([req_key]).values())[0]

	def get_daemon_cfg_options(self, req_keys):

		fn = self.get_daemon_cfg_fn()
		try:
			lines = get_lines_from_file(
				self.cfg, fn, desc='daemon config file', silent=not self.cfg.verbose)
		except:
			self.cfg._util.vmsg(f'Warning: {fn!r} does not exist or is unreadable')
			return dict((k, None) for k in req_keys)

		def gen():
			for key in req_keys:
				val = None
				for l in lines:
					if l.startswith(key):
						res = l.split('=', 1)
						if len(res) == 2 and not ' ' in res[1].strip():
							val = res[1].strip()
				yield (key, val)

		return dict(gen())

	def get_daemon_auth_cookie(self):
		fn = self.daemon.auth_cookie_fn
		return get_lines_from_file(
			self.cfg, fn, desc='cookie', quiet=True)[0] if os.access(fn, os.R_OK) else ''

	def info(self, info_id):

		def segwit_is_active():

			if 'deployment_info' in self.caps:
				return (
					self.cached['deploymentinfo']['deployments']['segwit']['active']
					or (self.cfg.test_suite and not self.chain == 'regtest')
				)

			d = self.cached['blockchaininfo']

			try:
				if d['softforks']['segwit']['active'] is True:
					return True
			except:
				pass

			try:
				if d['bip9_softforks']['segwit']['status'] == 'active':
					return True
			except:
				pass

			if self.cfg.test_suite and not self.chain == 'regtest':
				return True

			return False

		return locals()[info_id]()

	rpcmethods = (
		'backupwallet',
		'createrawtransaction',
		'decoderawtransaction',
		'disconnectnode',
		'estimatefee',
		'estimatesmartfee',
		'getaddressesbyaccount',
		'getaddressesbylabel',
		'getblock',
		'getblockchaininfo',
		'getblockcount',
		'getblockhash',
		'getblockheader',
		'getblockstats', # mmgen-node-tools
		'getdeploymentinfo',
		'getmempoolentry',
		'getmempoolentry',
		'getmempoolinfo',
		'getnettotals',
		'getnetworkinfo',
		'getpeerinfo',
		'getrawmempool',
		'getrawtransaction',
		'getrawtransaction',
		'gettransaction',
		'gettransaction',
		'getwalletinfo',
		'importaddress', # address (address or script) label rescan p2sh (Add P2SH version of the script)
		'importdescriptors', # like above, but for descriptor wallets
		'listaccounts',
		'listlabels',
		'listreceivedbylabel',
		'listsinceblock',
		'listunspent',
		'sendrawtransaction',
		'setlabel',
		'signrawtransaction',
		'signrawtransactionwithkey', # method new to Core v0.17.0
		'validateaddress',
		'walletpassphrase',
	)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
test.include.coin_daemon_control: Start and stop daemons for the MMGen test suite
"""

import sys
from pathlib import PurePath

sys.path[0] = str(PurePath(*PurePath(__file__).parts[:-3]))

from mmgen.cfg import Config, gc
from mmgen.util import msg, die, oneshot_warning
from mmgen.protocol import init_proto
from mmgen.daemon import CoinDaemon

xmr_wallet_network_ids = {
	'xmrw':    'mainnet',
	'xmrw_tn': 'testnet'
}

action = gc.prog_name.split('-')[0]

opts_data = {
	'sets': [('debug', True, 'verbose', True)],
	'text': {
		'desc': f'{action.capitalize()} coin or wallet daemons for the MMGen test suite',
		'usage':'[opts] <network IDs>',
		'options': """
-h, --help           Print this help message
--, --longhelp       Print help message for long (global) options
-D, --debug          Produce debugging output (implies --verbose)
-d, --datadir=       Override the default datadir
-i, --daemon-ids     Print all known daemon IDs
-m, --mainnet-only   Perform operations for mainnet daemons only
-n, --no-daemonize   Don't fork daemon to background
-p, --port-shift=    Shift the RPC port by this number
-r, --remove-datadir Remove the datadir(s) after stopping the daemon(s)
-s, --get-state      Get the state of the daemon(s) and exit
-t, --testing        Testing mode.  Print commands but don't execute them
-q, --quiet          Produce quieter output
-u, --usermode       Run the daemon in user (non test-suite) mode
-v, --verbose        Produce more verbose output
-V, --print-version  Print version strings from exec’ed daemons (not RPC)
-W, --no-wait        Don't wait for daemons to change state before exiting
""",
	'notes': """
Valid network IDs: {nid}, {xmrw_nid}, all, no_xmr
"""
	},
	'code': {
		'options': lambda s: s.format(
			a  = action.capitalize(),
			pn = gc.prog_name),
		'notes': lambda s, help_notes: s.format(
			nid = help_notes('coin_daemon_network_ids'),
			xmrw_nid = ', '.join(xmr_wallet_network_ids),
		)
	}
}

class warn_missing_exec(oneshot_warning):
	color = 'yellow'
	message = 'missing executable {!r}'

def run(network_id=None, proto=None, daemon_id=None, missing_exec_ok=False):

	if network_id in xmr_wallet_network_ids:
		from mmgen.proto.xmr.daemon import MoneroWalletDaemon
		d = MoneroWalletDaemon(
			cfg           = cfg,
			proto         = init_proto(cfg, coin='XMR', network=xmr_wallet_network_ids[network_id]),
			user          = 'test',
			passwd        = 'test passwd',
			test_suite    = True,
			monerod_addr  = None,
			trust_monerod = True,
			test_monerod  = False,
			opts          = ['no_daemonize'] if cfg.no_daemonize else None)
	else:
		d = CoinDaemon(
			cfg,
			network_id = network_id,
			proto      = proto,
			test_suite = not cfg.usermode,
			opts       = ['no_daemonize'] if cfg.no_daemonize else None,
			port_shift = int(cfg.port_shift or 0),
			datadir    = cfg.datadir,
			daemon_id  = daemon_id)

	if cfg.mainnet_only and d.network != 'mainnet':
		return

	d.debug = d.debug or cfg.debug
	d.wait = cfg.remove_datadir or not cfg.no_wait

	if missing_exec_ok:
		try:
			d.get_exec_version_str()
		except Exception as e:
			if not cfg.quiet:
				if cfg.verbose:
					msg(str(e))
				warn_missing_exec(div=d.exec_fn, fmt_args=(d.exec_fn,))
			return
	if cfg.print_version:
		msg('{:16} {}'.format(d.exec_fn+':', d.get_exec_version_str()))
	elif cfg.get_state:
		print(d.state_msg())
	elif cfg.testing:
		for cmd in d.start_cmds if action == 'start' else [d.stop_cmd]:
			print(' '.join(cmd))
	else:
		d.cmd(action, quiet=cfg.quiet)
		if action == 'stop' and cfg.remove_datadir:
			cfg._util.vmsg(f'Removing ‘{d.datadir}’')
			d.remove_datadir()

def main():

	if cfg.daemon_ids:
		print('\n'.join(CoinDaemon.all_daemon_ids()))
	elif 'all' in cfg._args or 'no_xmr' in cfg._args:
		if len(cfg._args) != 1:
			die(1, "'all' or 'no_xmr' must be the sole argument")
		for coin in CoinDaemon.coins:
			if coin == 'XMR' and cfg._args[0] == 'no_xmr':
				continue
			for daemon_id in CoinDaemon.get_daemon_ids(cfg, coin):
				for network in CoinDaemon.get_daemon(cfg, coin, daemon_id).networks:
					run(
						proto           = init_proto(cfg, coin=coin, network=network),
						daemon_id       = daemon_id,
						missing_exec_ok = True)
	else:
		ids = cfg._args
		network_ids = CoinDaemon.get_network_ids(cfg)
		if not ids:
			cfg._usage()
		for i in ids:
			if i not in network_ids + list(xmr_wallet_network_ids):
				die(1, f'{i!r}: invalid network ID')
		for network_id in ids:
			run(network_id=network_id.lower())

cfg = Config(opts_data=opts_data, init_opts={'skip_cfg_file': True})



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test/gentest.py: Cryptocoin key/address generation tests for the MMGen suite
"""

import sys, os, time

try:
	from include import test_init
except ImportError:
	from test.include import test_init

# Import these _after_ local path's been added to sys.path
from mmgen.cfg import gc, Config
from mmgen.color import green, red, purple
from mmgen.util import msg, ymsg, capfirst, is_int, die

results_file = 'gentest.out.json'

rounds = 100
opts_data = {
	'text': {
		'desc': 'Test key/address generation of the MMGen suite in various ways',
		'usage':'[options] <spec> <rounds | dump file>',
		'options': """
-h, --help         Print this help message
--, --longhelp     Print help message for long (global) options
-a, --all-coins    Test all coins supported by specified external tool
-k, --use-internal-keccak-module Force use of the internal keccak module
-q, --quiet        Produce quieter output
-s, --save-results Save output of external tool in Compare test to
                   {rf!r}
-t, --type=t       Specify address type (e.g. 'compressed', 'segwit',
                   'zcash_z', 'bech32')
-v, --verbose      Produce more verbose output
""",
	'notes': """
TEST TYPES:

  Compare: {prog} A:B <rounds>  (compare address generators A and B)
  Speed:   {prog} A <rounds>    (test speed of generator A)
  Dump:    {prog} A <dump file> (compare generator A to wallet dump)

  where:

      A and B are keygen backend numbers ('1' being the default); or
      B is the name of an external tool (see below) or 'ext'.

      If B is 'ext', the external tool will be chosen automatically.

  For the Compare test, A may be 'all' to test all backends for the current
  coin/address type combination.

EXAMPLES:

  Compare addresses generated by 'libsecp256k1' and 'python-ecdsa' backends,
  with 100 random rounds plus private-key edge cases:
  $ {prog} 1:2 100

  Compare Segwit addresses from default 'libsecp256k1' backend to 'pycoin'
  library for all supported coins, 100 rounds + edge cases:
  $ {prog} --all-coins --type=segwit 1:pycoin 100

  Compare addresses from 'python-ecdsa' backend to output of 'keyconv' tool
  for all supported coins, 100 rounds + edge cases:
  $ {prog} --all-coins --type=compressed 2:keyconv 100

  Compare bech32 addrs from 'libsecp256k1' backend to Bitcoin Core wallet
  dump:
  $ {prog} --type=bech32 1 bech32wallet.dump

  Compare addresses from Monero 'ed25519ll' backend to output of default
  external tool, 10 rounds + edge cases:
  $ {prog} --coin=xmr 3:ext 10

  Test the speed of default Monero 'nacl' backend, 10,000 rounds:
  $ test/gentest.py --coin=xmr 1 10000

  Same for Zcash:
  $ test/gentest.py --coin=zec --type=zcash_z 1 10000

  Test all configured Monero backends against the 'monero-python' library, 3 rounds
  + edge cases:
  $ test/gentest.py --coin=xmr all:monero-python 3

  Test 'nacl' and 'ed25519ll_djbec' backends against each other, 10,000 rounds
  + edge cases:
  $ test/gentest.py --coin=xmr 1:2 10000

SUPPORTED EXTERNAL TOOLS:

  + eth-keys (for ETH, ETC)
    https://github.com/ethereum/eth-keys

  + ethkey (eth-keys alternative for MSYS2, or if eth-keys is unavailable)
    https://github.com/openethereum/openethereum/releases/tag/v3.1.0

  + zcash-mini (for Zcash-Z addresses and view keys)
    https://github.com/FiloSottile/zcash-mini

  + monero-python (for Monero addresses and view keys)
    https://github.com/monero-ecosystem/monero-python

  + pycoin (for supported coins)
    https://github.com/richardkiss/pycoin

  + keyconv (for supported coins)
    https://github.com/exploitagency/vanitygen-plus
    ('keyconv' does not generate Segwit addresses)
"""
	},
	'code': {
		'options': lambda s: s.format(
			rf   = results_file,
		),
		'notes': lambda s: s.format(
			prog = 'test/gentest.py',
			pnm  = gc.proj_name,
			snum = rounds)
	}
}

def get_cmd_output(cmd, input=None):
	return run(cmd, input=input, stdout=PIPE, stderr=DEVNULL).stdout.decode().splitlines()

saved_results = {}

class GenTool:

	def __init__(self, proto, addr_type):
		self.proto = proto
		self.addr_type = addr_type
		self.data = {}

	def __del__(self):
		if cfg.save_results:
			key = f'{self.proto.coin}-{self.proto.network}-{self.addr_type.name}-{self.desc}'.lower()
			saved_results[key] = {k.hex():v._asdict() for k, v in self.data.items()}

	def run_tool(self, sec, cache_data):
		vcoin = 'BTC' if self.proto.coin == 'BCH' else self.proto.coin
		key = sec.orig_bytes
		if key in self.data:
			return self.data[key]
		else:
			ret = self.run(sec, vcoin)
			if cache_data:
				self.data[key] = sd(**{'reduced':sec.hex()}, **ret._asdict())
			return ret

class GenToolEth_keys(GenTool):
	desc = 'eth-keys'

	def __init__(self, *args, **kwargs):
		self.keys = self.cmdname = None
		try:
			from eth_keys import keys
			self.keys = keys
		except ImportError:
			self.cmdname = get_ethkey()
			self.desc = 'ethkey'
		if not (self.keys or self.cmdname):
			die(2, 'Neither the ‘eth-keys’ package nor the ‘ethkey’ executable '
				'could be found on the system!')
		super().__init__(*args, **kwargs)

	def run(self, sec, vcoin):
		if self.keys:
			sk = self.keys.PrivateKey(sec)
			return gtr(str(sk)[2:], sk.public_key.to_address()[2:], None)
		else:
			o = get_cmd_output([self.cmdname, 'info', sec.hex()])
			return gtr(o[0].split()[1], o[-1].split()[1], None)

class GenToolKeyconv(GenTool):
	desc = 'keyconv'
	def run(self, sec, vcoin):
		o = get_cmd_output(['keyconv', '-C', vcoin, sec.wif])
		return gtr(
			o[1].split()[1],
			o[0].split()[1],
			None)

class GenToolZcash_mini(GenTool):
	desc = 'zcash-mini'
	def run(self, sec, vcoin):
		o = get_cmd_output(['zcash-mini', '-key', '-simple'], input=(sec.wif+'\n').encode())
		return gtr(o[1], o[0], o[-1])

class GenToolPycoin(GenTool):
	"""
	pycoin/networks/all.py pycoin/networks/legacy_networks.py
	"""
	desc = 'pycoin'
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		try:
			from pycoin.networks.registry import network_for_netcode
		except ImportError as e:
			raise ImportError(
				f'{e}\nUnable to import pycoin.networks.registry. Is pycoin installed on your system?') from e
		self.nfnc = network_for_netcode

	def run(self, sec, vcoin):
		if self.proto.testnet:
			vcoin = cinfo.external_tests['testnet']['pycoin'][vcoin]
		network = self.nfnc(vcoin)
		key = network.keys.private(
			secret_exponent = int(sec.hex(), 16),
			is_compressed = self.addr_type.name != 'legacy')
		if key is None:
			die(1, f'can’t parse {sec.hex()}')
		if self.addr_type.name in ('segwit', 'bech32'):
			hash160_c = key.hash160(is_compressed=True)
			if self.addr_type.name == 'segwit':
				p2sh_script = network.contract.for_p2pkh_wit(hash160_c)
				addr = network.address.for_p2s(p2sh_script)
			else:
				addr = network.address.for_p2pkh_wit(hash160_c)
		else:
			addr = key.address()
		return gtr(key.wif(), addr, None)

class GenToolMonero_python(GenTool):
	desc = 'monero-python'

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		try:
			from monero.seed import Seed
		except ImportError as e:
			raise ImportError(
				f'{e}\nUnable to import monero-python. Is monero-python installed on your system?') from e
		self.Seed = Seed

	def run(self, sec, vcoin):
		seed = self.Seed(sec.orig_bytes.hex())
		sk = seed.secret_spend_key()
		vk = seed.secret_view_key()
		addr = seed.public_address()
		return gtr(sk, addr, vk)

def find_or_check_tool(proto, addr_type, toolname):

	ext_progs = list(cinfo.external_tests[proto.network])

	if toolname not in ext_progs + ['ext']:
		die(1, f'{toolname!r}: unsupported tool for network {proto.network}')

	if cfg.all_coins and toolname == 'ext':
		die(1, "'--all-coins' must be combined with a specific external testing tool")
	else:
		tool = cinfo.get_test_support(
			proto.coin,
			addr_type.name,
			proto.network,
			verbose = not cfg.quiet,
			toolname = toolname if toolname != 'ext' else None)
		if tool and toolname in ext_progs and toolname != tool:
			sys.exit(3)
		if tool is None:
			return None
		return tool

def test_equal(desc, a_val, b_val, in_bytes, sec, wif, a_desc, b_desc):
	if a_val != b_val:
		fs = """
  {i:{w}}: {}
  {s:{w}}: {}
  {W:{w}}: {}
  {a:{w}}: {}
  {b:{w}}: {}
			"""
		die(3,
			red('\nERROR: {} do not match!').format(desc)
			+ fs.format(
				in_bytes.hex(), sec, wif, a_val, b_val,
				i='input', s='sec key', W='WIF key', a=a_desc, b=b_desc,
				w=max(len(e) for e in (a_desc, b_desc)) + 1
		).rstrip())

def do_ab_test(proto, scfg, addr_type, gen1, kg2, ag, tool, cache_data):

	def do_ab_inner(n, trounds, in_bytes):
		global last_t
		if cfg.verbose or time.time() - last_t >= 0.1:
			qmsg_r(f'\rRound {i+1}/{trounds} ')
			last_t = time.time()
		sec = PrivKey(proto, in_bytes, compressed=addr_type.compressed, pubkey_type=addr_type.pubkey_type)
		data = kg1.gen_data(sec)
		addr1 = ag.to_addr(data)
		view_pref = 1 if proto.coin == 'BCH' else 0
		tinfo = (in_bytes, sec, sec.wif, type(kg1).__name__, type(kg2).__name__ if kg2 else tool.desc)

		def do_msg():
			if cfg.verbose:
				msg(fs.format(b=in_bytes.hex(), r=sec.hex(), k=sec.wif, v=vk2, a=addr1))

		if tool:
			def run_tool():
				o = tool.run_tool(sec, cache_data)
				test_equal('WIF keys', sec.wif, o.wif, *tinfo)
				test_equal('addresses', addr1.views[view_pref], o.addr, *tinfo)
				if o.viewkey:
					test_equal('view keys', ag.to_viewkey(data), o.viewkey, *tinfo)
				return o.viewkey
			vk2 = run_tool()
			do_msg()
		else:
			test_equal('addresses', addr1.views[view_pref], ag.to_addr(kg2.gen_data(sec)), *tinfo)
			vk2 = None
			do_msg()

		qmsg_r(f'\rRound {n+1}/{trounds} ')

	def get_randbytes():
		if tool and len(tool.data) > len(edgecase_sks):
			yield from tuple(tool.data)[len(edgecase_sks):]
		else:
			for _ in range(scfg.rounds):
				yield getrand(32)

	kg1 = KeyGenerator(cfg, proto, addr_type.pubkey_type, backend=gen1)
	if type(kg1) == type(kg2):
		die(4, 'Key generators are the same!')

	e = cinfo.get_entry(proto.coin, proto.network)
	qmsg(green("Comparing address generators '{A}' and '{B}' for {N} {c} ({n}), addrtype {a!r}".format(
		A = type(kg1).__name__.replace('_', '-'),
		B = type(kg2).__name__.replace('_', '-') if kg2 else tool.desc,
		N = proto.network,
		c = proto.coin,
		n = e.name if e else '---',
		a = addr_type.name)))

	global last_t
	last_t = time.time()

	fs  = (
		'\ninput:    {b}' +
		'\nreduced:  {r}' +
		'\n{:9} {{k}}'.format(addr_type.wif_label+':') +
		('\nviewkey:  {v}' if 'viewkey' in addr_type.extra_attrs else '') +
		'\naddr:     {a}\n')

	group_order = CoinProtocol.Secp256k1.secp256k1_group_order

	# test some important private key edge cases:
	edgecase_sks = (
		bytes([0x00]*31 + [0x01]), # min
		bytes([0xff]*32),          # max
		bytes([0x0f] + [0xff]*31), # produces same key as above for zcash-z
		int.to_bytes(group_order + 1, 32, 'big'), # bitcoin will reduce
		int.to_bytes(group_order - 1, 32, 'big'), # bitcoin will not reduce
		bytes([0x00]*31 + [0xff]), # monero will reduce
		bytes([0xff]*31 + [0x0f]), # monero will not reduce
		bytes.fromhex('deadbeef'*8),
	)

	qmsg(purple('edge cases:'))
	for i, privbytes in enumerate(edgecase_sks):
		do_ab_inner(i, len(edgecase_sks), privbytes)
	qmsg(green('\rOK            ' if cfg.verbose else 'OK'))

	qmsg(purple('random input:'))
	for i, privbytes in enumerate(get_randbytes()):
		do_ab_inner(i, scfg.rounds, privbytes)
	qmsg(green('\rOK            ' if cfg.verbose else 'OK'))

def init_tool(proto, addr_type, toolname):
	return globals()['GenTool'+capfirst(toolname.replace('-', '_'))](proto, addr_type)

def ab_test(proto, scfg):

	addr_type = MMGenAddrType(proto=proto, id_str=cfg.type or proto.dfl_mmtype)

	if scfg.gen2:
		assert scfg.gen1 != 'all', "'all' must be used only with external tool"
		kg2 = KeyGenerator(cfg, proto, addr_type.pubkey_type, backend=scfg.gen2)
		tool = None
	else:
		toolname = find_or_check_tool(proto, addr_type, scfg.tool)
		if toolname is None:
			ymsg(f'Warning: skipping tool {scfg.tool!r} for {proto.coin} {addr_type.name}')
			return
		tool = init_tool(proto, addr_type, toolname)
		kg2 = None

	ag = AddrGenerator(cfg, proto, addr_type)

	if scfg.all_backends: # check all backends against external tool
		for n in range(len(get_backends(addr_type.pubkey_type))):
			do_ab_test(
					proto,
					scfg,
					addr_type,
					gen1       = n+1,
					kg2        = kg2,
					ag         = ag,
					tool       = tool,
					cache_data = scfg.rounds < 1000 and not n)
	else:                # check specific backend against external tool or another backend
		do_ab_test(
				proto,
				scfg,
				addr_type,
				gen1       = scfg.gen1,
				kg2        = kg2,
				ag         = ag,
				tool       = tool,
				cache_data = False)

def speed_test(proto, kg, ag, rounds):
	qmsg(green('Testing speed of address generator {!r} for coin {}'.format(
		type(kg).__name__,
		proto.coin)))
	from struct import pack
	seed = getrand(28)
	qmsg('Incrementing key with each round')
	qmsg('Starting key: {}'.format((seed + pack('I', 0)).hex()))
	start = last_t = time.time()

	for i in range(rounds):
		if time.time() - last_t >= 0.1:
			qmsg_r(f'\rRound {i+1}/{rounds} ')
			last_t = time.time()
		sec = PrivKey(proto, seed+pack('I', i), compressed=ag.compressed, pubkey_type=ag.pubkey_type)
		addr = ag.to_addr(kg.gen_data(sec))
		vmsg(f'\nkey:  {sec.wif}\naddr: {addr}\n')
	qmsg(
		f'\rRound {i+1}/{rounds} ' +
		f'\n{rounds} addresses generated' +
		('' if cfg.test_suite_deterministic else f' in {time.time()-start:.2f} seconds')
	)

def dump_test(proto, kg, ag, filename):

	with open(filename) as fp:
		dump = [[*(e.split()[0] for e in line.split('addr='))] for line in fp.readlines() if 'addr=' in line]
		if not dump:
			die(1, f'File {filename!r} appears not to be a wallet dump')

	qmsg(green(
		"A: generator pair '{}:{}'\nB: wallet dump {!r}".format(
			type(kg).__name__,
			type(ag).__name__,
			filename)))

	for count, (b_wif, b_addr) in enumerate(dump, 1):
		qmsg_r(f'\rKey {count}/{len(dump)} ')
		try:
			b_sec = PrivKey(proto, wif=b_wif)
		except:
			die(2, f'\nInvalid {proto.network} WIF address in dump file: {b_wif}')
		a_addr = ag.to_addr(kg.gen_data(b_sec))
		vmsg(f'\nwif: {b_wif}\naddr: {b_addr}\n')
		tinfo = (b_sec, b_sec.hex(), b_wif, type(kg).__name__, filename)
		test_equal('addresses', a_addr, b_addr, *tinfo)

	qmsg(green(('\n', '')[bool(cfg.verbose)] + 'OK'))

def get_protos(proto, addr_type, toolname):

	init_genonly_altcoins(testnet=proto.testnet)

	for coin in cinfo.external_tests[proto.network][toolname]:
		if coin.lower() not in CoinProtocol.coins:
			continue
		ret = init_proto(cfg, coin, testnet=proto.testnet)
		if addr_type not in ret.mmtypes:
			continue
		yield ret

def parse_args():

	if len(cfg._args) != 2:
		cfg._usage()

	arg1, arg2 = cfg._args
	gen1, gen2, rounds = (0, 0, 0)
	tool, all_backends, dumpfile = (None, None, None)

	if is_int(arg1) and is_int(arg2):
		test = 'speed'
		gen1 = arg1
		rounds = arg2
	elif is_int(arg1) and os.access(arg2, os.R_OK):
		test = 'dump'
		gen1 = arg1
		dumpfile = arg2
	else:
		test = 'ab'
		rounds = arg2

		if not is_int(arg2):
			die(1, 'Second argument must be dump filename or integer rounds specification')

		try:
			a, b = arg1.split(':')
		except:
			die(1, 'First argument must be a generator backend number or two colon-separated arguments')

		if is_int(a):
			gen1 = a
		else:
			if a == 'all':
				all_backends = True
			else:
				die(1, "First part of first argument must be a generator backend number or 'all'")

		if is_int(b):
			if cfg.all_coins:
				die(1, '--all-coins must be used with external tool only')
			gen2 = b
		else:
			tool = b
			ext_progs = list(cinfo.external_tests[cfg._proto.network]) + ['ext']
			if b not in ext_progs:
				die(1, f'Second part of first argument must be a generator backend number or one of {ext_progs}')

	return namedtuple('parsed_args',
			['test', 'gen1', 'gen2', 'rounds', 'tool', 'all_backends', 'dumpfile'])(
		test,
		int(gen1) or None,
		int(gen2) or None,
		int(rounds) or None,
		tool,
		all_backends,
		dumpfile)

def main():

	scfg = parse_args()

	addr_type = MMGenAddrType(proto=proto, id_str=cfg.type or proto.dfl_mmtype)

	if scfg.test == 'ab':
		protos = get_protos(proto, addr_type, scfg.tool) if cfg.all_coins else [proto]
		for p in protos:
			ab_test(p, scfg)
	else:
		kg = KeyGenerator(cfg, proto, addr_type.pubkey_type, backend=scfg.gen1)
		ag = AddrGenerator(cfg, proto, addr_type)
		if scfg.test == 'speed':
			speed_test(proto, kg, ag, scfg.rounds)
		elif scfg.test == 'dump':
			dump_test(proto, kg, ag, scfg.dumpfile)

	if saved_results:
		import json
		with open(results_file, 'w') as fp:
			fp.write(json.dumps(saved_results, indent=4))

from subprocess import run, PIPE, DEVNULL
from collections import namedtuple
from mmgen.protocol import init_proto, CoinProtocol
from mmgen.altcoin.params import init_genonly_altcoins
from test.altcointest import TestCoinInfo as cinfo
from mmgen.key import PrivKey
from mmgen.addr import MMGenAddrType
from mmgen.addrgen import KeyGenerator, AddrGenerator
from mmgen.keygen import get_backends
from test.include.common import getrand, get_ethkey, set_globals

gtr = namedtuple('gen_tool_result', ['wif', 'addr', 'viewkey'])
sd = namedtuple('saved_data_item', ['reduced', 'wif', 'addr', 'viewkey'])

sys.argv = [sys.argv[0]] + ['--skip-cfg-file'] + sys.argv[1:]

cfg = Config(opts_data=opts_data)
set_globals(cfg)

qmsg = cfg._util.qmsg
qmsg_r = cfg._util.qmsg_r
vmsg = cfg._util.vmsg

proto = cfg._proto

if proto.coin in ('XMR', 'ETH', 'ETC'):
	from mmgen.util2 import load_cryptodome
	load_cryptodome()

if __name__ == '__main__':
	from mmgen.main import launch
	launch(func=main)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
tw.view: base class for tracking wallet view classes
"""

import sys, time, asyncio
from collections import namedtuple

from ..cfg import gv
from ..objmethods import MMGenObject
from ..obj import get_obj, MMGenIdx, MMGenList
from ..color import nocolor, yellow, orange, green, red, blue
from ..util import msg, msg_r, fmt, die, capfirst, suf, make_timestr
from ..rpc import rpc_init
from ..base_obj import AsyncInit

# these are replaced by fake versions in overlay:
CUR_HOME  = '\033[H'
CUR_UP    = lambda n: f'\033[{n}A'
CUR_DOWN  = lambda n: f'\033[{n}B'
ERASE_ALL = '\033[0J'

# decorator for action.run():
def enable_echo(orig_func):
	async def f(self, parent, action_method):
		if parent.scroll:
			parent.term.set('echo')
		ret = await orig_func(self, parent, action_method)
		if parent.scroll:
			parent.term.set('noecho')
		return ret
	return f

# base class for TwUnspentOutputs, TwAddresses, TwTxHistory:
class TwView(MMGenObject, metaclass=AsyncInit):

	class display_type:

		class squeezed:
			detail = False
			fmt_method = 'gen_squeezed_display'
			line_fmt_method = 'squeezed_format_line'
			subhdr_fmt_method = 'gen_subheader'
			colhdr_fmt_method = 'squeezed_col_hdr'
			need_column_widths = True
			item_separator = '\n'
			print_header = '[screen print truncated to width {}]\n'

		class detail:
			detail = True
			fmt_method = 'gen_detail_display'
			line_fmt_method = 'detail_format_line'
			subhdr_fmt_method = 'gen_subheader'
			colhdr_fmt_method = 'detail_col_hdr' # set to None to disable
			need_column_widths = True
			item_separator = '\n'
			print_header = ''

	class line_processing:

		class print:
			@staticmethod
			def do(method, data, cw, fs, color, fmt_method):
				return [l.rstrip() for l in method(data, cw, fs, color, fmt_method)]

	has_wallet  = True
	has_amt2    = False
	dates_set   = False
	reverse     = False
	group       = False
	use_cached  = False
	minconf     = 1
	txid_w      = 0
	txid_max_w  = {}
	txid_min_w  = {}
	txid_nice_w = {}
	nice_addr_w = {'addr': 14}
	sort_key    = 'age'
	display_hdr = ()
	display_body = ()
	prompt_fs_repl = {}
	nodata_msg = '[no data for requested parameters]'
	cols = 0
	term_height = 0
	term_width = 0
	scrollable_height = 0
	min_scrollable_height = 5
	addr_view_pref = 0
	pos = 0
	filters = ()

	fp = namedtuple('fs_params', ['fs_key', 'hdr_fs_repl', 'fs_repl', 'hdr_fs', 'fs'])
	fs_params = {
		'num':       fp('n', True, True,  ' {n:>%s}', ' {n:>%s}'),
		'txid':      fp('t', True, False, ' {t:%s}',  ' {t}'),
		'vout':      fp('v', True, False, '{v:%s}',   '{v}'),
		'used':      fp('u', True, False, ' {u:%s}',  ' {u}'),
		'addr':      fp('a', True, False, ' {a:%s}',  ' {a}'),
		'mmid':      fp('m', True, False, ' {m:%s}',  ' {m}'),
		'comment':   fp('c', True, False, ' {c:%s}',  ' {c}'),
		'amt':       fp('A', True, False, ' {A:%s}',  ' {A}'),
		'amt2':      fp('B', True, False, ' {B:%s}',  ' {B}'),
		'date':      fp('d', True, True,  ' {d:%s}',  ' {d:<%s}'),
		'date_time': fp('D', True, True,  ' {D:%s}',  ' {D:%s}'),
		'block':     fp('b', True, True,  ' {b:%s}',  ' {b:<%s}'),
		'inputs':    fp('i', True, False, ' {i:%s}',  ' {i}'),
		'outputs':   fp('o', True, False, ' {o:%s}',  ' {o}'),
	}

	age_fmts = ('confs', 'block', 'days', 'date', 'date_time')
	age_fmts_date_dependent = ('days', 'date', 'date_time')
	_age_fmt = 'confs'

	bch_addr_fmts = ('cashaddr', 'legacy')

	age_col_params = {
		'confs':     (0,  'Confs'),
		'block':     (0,  'Block'),
		'days':      (0,  'Age(d)'),
		'date':      (0,  'Date'),
		'date_time': (0,  'Date/Time'),
	}

	date_formatter = {
		'days': lambda rpc, secs: (rpc.cur_date - secs) // 86400 if secs else 0,
		'date': (
			lambda rpc, secs: '{}-{:02}-{:02}'.format(*time.gmtime(secs)[:3])[2:]
				if secs else '-       '),
		'date_time': (
			lambda rpc, secs: '{}-{:02}-{:02} {:02}:{:02}'.format(*time.gmtime(secs)[:5])
				if secs else '-               '),
	}

	twidth_diemsg = """
		--columns or MMGEN_COLUMNS value ({}) is too small to display the {}
		Minimum value for this configuration: {}
	"""
	twidth_errmsg = """
		Screen is too narrow to display the {} with current configuration
		Please resize your screen to at least {} characters and hit any key:
	"""
	theight_errmsg = """
		Terminal window is too small to display the {} with current configuration
		Please resize it to at least {} lines and hit any key:
	"""

	squeezed_format_line = None
	detail_format_line = None

	scroll_keys = {
		'vi': {
			'k': 'm_cursor_up',
			'j': 'm_cursor_down',
			'b': 'm_pg_up',
			'f': 'm_pg_down',
			'g': 'm_top',
			'G': 'm_bot',
		},
		'linux': {
			'\x1b[A': 'm_cursor_up',
			'\x1b[B': 'm_cursor_down',
			'\x1b[5~': 'm_pg_up',
			'\x1b[6~': 'm_pg_down',
			'\x1b[7~': 'm_top',
			'\x1b[8~': 'm_bot',
		},
		'win32': {
			'\xe0H': 'm_cursor_up',
			'\xe0P': 'm_cursor_down',
			'\xe0I': 'm_pg_up',
			'\xe0Q': 'm_pg_down',
			'\xe0G': 'm_top',
			'\xe0O': 'm_bot',
		}
	}
	scroll_keys['darwin'] = scroll_keys['linux']
	extra_key_mappings = {}

	def __new__(cls, cfg, proto, *args, **kwargs):
		return MMGenObject.__new__(proto.base_proto_subclass(cls, cls.mod_subpath))

	async def __init__(self, cfg, proto):
		self.cfg = cfg
		self.proto = proto
		self.rpc = await rpc_init(cfg, proto)
		if self.has_wallet:
			from .ctl import TwCtl
			self.twctl = await TwCtl(cfg, proto, mode='w')
		self.amt_keys = {'amt':'iwidth', 'amt2':'iwidth2'} if self.has_amt2 else {'amt':'iwidth'}
		if repl := self.prompt_fs_repl.get(self.proto.coin):
			self.prompt_fs_in[repl[0]] = repl[1]
		self.prompt_fs = '\n'.join(self.prompt_fs_in)
		self.key_mappings.update(self.extra_key_mappings)
		if self.proto.coin == 'BCH':
			self.key_mappings.update({'h': 'd_addr_view_pref'})
			self.addr_view_pref = 1 if not self.cfg.cashaddr else not self.proto.cashaddr

	@property
	def age_w(self):
		return self.age_col_params[self.age_fmt][0]

	@property
	def age_hdr(self):
		return self.age_col_params[self.age_fmt][1]

	@property
	def age_fmt(self):
		return self._age_fmt

	@age_fmt.setter
	def age_fmt(self, val):
		if val not in self.age_fmts:
			die('BadAgeFormat', f'{val!r}: invalid age format (must be one of {self.age_fmts!r})')
		self._age_fmt = val

	def age_disp(self, o, age_fmt):
		if self.has_age:
			if age_fmt == 'confs':
				return o.confs or '-'
			elif age_fmt == 'block':
				return self.rpc.blockcount + 1 - o.confs if o.confs else '-'
			else:
				return self.date_formatter[age_fmt](self.rpc, o.date)

	def get_disp_prec(self, wide):
		return self.proto.coin_amt.max_prec

	sort_disp = {
		'addr':   'Addr',
		'age':    'Age',
		'amt':    'Amt',
		'txid':   'TxID',
		'twmmid': 'MMGenID',
	}

	sort_funcs = {
		'addr':   lambda i: i.addr,
		'age':    lambda i: 0 - i.confs,
		'amt':    lambda i: i.amt,
		'txid':   lambda i: f'{i.txid} {i.vout:04}',
		'twmmid': lambda i: i.twmmid.sort_key
	}

	def sort_info(self, *, include_group=True):
		ret = ([], ['Reverse'])[self.reverse]
		ret.append(self.sort_disp[self.sort_key])
		if include_group and self.group and (self.sort_key in ('addr', 'txid', 'twmmid')):
			ret.append('Grouped')
		return ret

	def do_sort(self, key=None, *, reverse=False):
		if key == 'txid' and not self.txid_w:
			return
		key = key or self.sort_key
		if key not in self.sort_funcs:
			die(1, f'{key!r}: invalid sort key.  Valid options: {" ".join(self.sort_funcs)}')
		self.sort_key = key
		assert isinstance(reverse, bool)
		save = self.data.copy()
		self.data.sort(key=self.sort_funcs[key], reverse=reverse or self.reverse)
		if self.data != save:
			self.pos = 0

	async def get_data(self, *, sort_key=None, reverse_sort=False):

		rpc_data = await self.get_rpc_data()

		if not rpc_data:
			die(1, fmt(self.no_rpcdata_errmsg).strip())

		lbl_id = ('account', 'label')['label_api' in self.rpc.caps]

		res = self.gen_data(rpc_data, lbl_id)
		self.data = MMGenList(await res if type(res).__name__ == 'coroutine' else res)
		self.disp_data = list(self.filter_data())

		if not self.data:
			die(1, f'No {self.item_desc_pl} in tracking wallet!')

		self.do_sort(key=sort_key, reverse=reverse_sort)

		# get_data() is immediately followed by display header, and get_rpc_data() produces output,
		# so add NL here (' ' required because CUR_HOME erases preceding blank lines)
		msg(' ')

	def get_term_dimensions(self, min_cols, *, min_lines=None):
		from ..term import get_terminal_size, get_char_raw, _term_dimensions
		user_resized = False
		while True:
			ts = get_terminal_size()
			cols = self.cfg.columns or ts.width
			lines = ts.height
			if cols >= min_cols and (min_lines is None or lines >= min_lines):
				if user_resized:
					msg_r(CUR_HOME + ERASE_ALL)
				return _term_dimensions(cols, ts.height)
			if sys.stdout.isatty():
				if self.cfg.columns and cols < min_cols:
					die(1, '\n'+fmt(self.twidth_diemsg.format(self.cfg.columns, self.desc, min_cols), indent='  '))
				else:
					m, dim = (self.twidth_errmsg, min_cols) if cols < min_cols else (self.theight_errmsg, min_lines)
					get_char_raw(CUR_HOME + ERASE_ALL + fmt(m.format(self.desc, dim), append=''))
					user_resized = True
			else:
				return _term_dimensions(min_cols, ts.height)

	def compute_column_widths(self, widths, maxws, minws, maxws_nice, *, wide, interactive):

		def do_ret(freews):
			widths.update({k:minws[k] + freews.get(k, 0) for k in minws})
			widths.update({ikey: widths[key] - self.disp_prec - 1 for key, ikey in self.amt_keys.items()})
			return namedtuple('column_widths', widths.keys())(*widths.values())

		def do_ret_max():
			widths.update({k:max(minws[k], maxws[k]) for k in minws})
			widths.update({ikey: widths[key] - self.disp_prec - 1 for key, ikey in self.amt_keys.items()})
			return namedtuple('column_widths', widths.keys())(*widths.values())

		def get_freews(cols, varws, varw, minw):
			freew = cols - minw
			if freew and varw:
				x = freew / varw
				freews = {k:int(varws[k] * x) for k in varws}
				remainder = freew - sum(freews.values())
				for k in varws:
					if not remainder:
						break
					if freews[k] < varws[k]:
						freews[k] += 1
						remainder -= 1
				return freews
			else:
				return {k:0 for k in varws}

		varws = {k:maxws[k] - minws[k] for k in maxws if maxws[k] > minws[k]}
		minw = sum(widths.values()) + sum(minws.values())
		varw = sum(varws.values())

		self.min_term_width = 40 if wide else max(self.prompt_width, minw) if interactive else minw
		td = self.get_term_dimensions(self.min_term_width)
		self.term_height = td.height
		self.term_width = td.width

		self.cols = min(self.term_width, minw + varw)

		if wide or self.cols == minw + varw:
			return do_ret_max()

		if maxws_nice:
			# compute high-priority widths:
			varws_hp = {k: maxws_nice[k] - minws[k] if k in maxws_nice else varws[k] for k in varws}
			varw_hp = sum(varws_hp.values())
			widths_hp = get_freews(
				min(self.term_width, minw + varw_hp),
				varws_hp,
				varw_hp,
				minw)
			# compute low-priority (nice) widths:
			varws_lp = {k: varws[k] - varws_hp[k] for k in maxws_nice if k in varws}
			widths_lp = get_freews(
				self.cols,
				varws_lp,
				sum(varws_lp.values()),
				minw + sum(widths_hp.values()))
			# sum the two for each field:
			return do_ret({k:widths_hp[k] + widths_lp.get(k, 0) for k in varws})
		else:
			return do_ret(get_freews(self.cols, varws, varw, minw))

	def gen_subheader(self, cw, color):
		c_orange = (nocolor, orange)[color]
		c_yellow = (nocolor, yellow)[color]
		if self.rpc.is_remote:
			yield (
				c_orange(f'WARNING: Connecting to public {self.rpc.server_proto} node at ') +
				self.rpc.server_domain.hl(color=color))
			yield c_orange('         To improve anonymity, proxy requests via Tor or I2P')
		if self.twctl.use_cached_balances:
			yield c_yellow('Using cached balances. These may be out of date!')
		else:
			yield f'Displaying balances with {self.minconf} confirmation{suf(self.minconf)}'

	def gen_footer(self, color):
		if hasattr(self, 'total'):
			yield 'TOTAL: {} {}'.format(self.total.hl(color=color), self.proto.dcoin)

	def set_amt_widths(self, data):
		# width of amts column: min(7, width of integer part) + len('.') + width of fractional part
		self.amt_widths = {
			k:min(7, max(len(str(getattr(d, k).to_integral_value())) for d in data)) + 1 + self.disp_prec
				for k in self.amt_keys}

	async def format(
			self,
			display_type,
			*,
			color           = True,
			interactive     = False,
			line_processing = None,
			scroll          = False):

		def make_display():

			def gen_hdr(spc):

				Blue, Green = (blue, green) if color else (nocolor, nocolor)
				Yes, No, All = (green('yes'), red('no'), yellow('all')) if color else ('yes', 'no', 'all')
				sort_info = ' '.join(self.sort_info())

				def fmt_filter(k):
					return '{}:{}'.format(k, {0:No, 1:Yes, 2:All}[getattr(self, k)])

				yield '{} (sort order: {}){}'.format(
					self.hdr_lbl.upper(),
					Blue(sort_info),
					spc * (self.cols - len(f'{self.hdr_lbl} (sort order: {sort_info})')))

				if self.filters:
					yield 'Filters: {}{}'.format(
						' '.join(map(fmt_filter, self.filters)),
						spc * len(self.filters))

				yield 'Network: {}'.format(Green(
					self.proto.coin + ' ' + self.proto.chain_name.upper()))

				if not self.rpc.is_remote:
					yield 'Block {} [{}]'.format(
						self.rpc.blockcount.hl(color=color),
						make_timestr(self.rpc.cur_date))

				if hasattr(self, 'total'):
					yield 'Total {}: {}'.format(self.proto.dcoin, self.total.hl(color=color))

				yield from getattr(self, dt.subhdr_fmt_method)(cw, color)

				yield spc * self.term_width

				if data and dt.colhdr_fmt_method:
					col_hdr = getattr(self, dt.colhdr_fmt_method)(cw, hdr_fs, color)
					yield col_hdr.rstrip() if line_processing == 'print' else col_hdr

			def get_body(method):
				if line_processing:
					return getattr(self.line_processing, line_processing).do(
						method, data, cw, fs, color, getattr(self, dt.line_fmt_method))
				else:
					return method(data, cw, fs, color, getattr(self, dt.line_fmt_method))

			if data and dt.need_column_widths:
				self.set_amt_widths(data)
				cw = self.get_column_widths(data, wide=dt.detail, interactive=interactive)
				cwh = cw._asdict()
				fp = self.fs_params
				rfill = ' ' * (self.term_width - self.cols) if scroll else ''
				hdr_fs = ''.join(fp[name].hdr_fs % ((), cwh[name])[fp[name].hdr_fs_repl]
					for name in dt.cols if cwh[name]) + rfill
				fs = ''.join(fp[name].fs % ((), cwh[name])[fp[name].fs_repl]
					for name in dt.cols if cwh[name]) + rfill
			else:
				cw = hdr_fs = fs = None

			return (
				tuple(gen_hdr(spc='' if line_processing == 'print' else ' ')),
				tuple(
					get_body(getattr(self, dt.fmt_method)) if data else
					[(nocolor, yellow)[color](self.nodata_msg.ljust(self.term_width))])
			)

		if not gv.stdout.isatty():
			line_processing = 'print'

		dt = getattr(self.display_type, display_type)

		if self.use_cached:
			self.use_cached = False
			display_hdr = self.display_hdr
			display_body = self.display_body
		else:
			self.disp_prec = self.get_disp_prec(wide=dt.detail)

			if self.has_age and (self.age_fmt in self.age_fmts_date_dependent or dt.detail):
				await self.set_dates(self.data)

			dsave = self.disp_data
			data = self.disp_data = list(self.filter_data()) # method could be a generator

			if data != dsave:
				self.pos = 0

			display_hdr, display_body = make_display()

			if scroll:
				fixed_height = len(display_hdr) + self.prompt_height + 1

				if self.term_height - fixed_height < self.min_scrollable_height:
					td = self.get_term_dimensions(
						self.min_term_width,
						min_lines = self.min_scrollable_height + fixed_height)
					self.term_height = td.height
					self.term_width = td.width
					display_hdr, display_body = make_display()

				self.scrollable_height = self.term_height - fixed_height
				self.max_pos = max(0, len(display_body) - self.scrollable_height)
				self.pos = min(self.pos, self.max_pos)

			if not dt.detail:
				self.display_hdr = display_hdr
				self.display_body = display_body

		if scroll:
			top = self.pos
			bot = self.pos + self.scrollable_height
			fill = ('\n' + ''.ljust(self.term_width)) * (self.scrollable_height - len(display_body))
		else:
			top, bot, fill = (None, None, '')

		if interactive:
			footer = ''
		else:
			footer = '\n'.join(self.gen_footer(color))
			footer = ('\n\n' + footer if footer else '') + '\n'

		return (
			'\n'.join(display_hdr) + '\n'
			+ dt.item_separator.join(display_body[top:bot])
			+ fill
			+ footer
		)

	async def view_filter_and_sort(self):

		action_map = {
			'a_': 'action',
			's_': 'sort_action',
			'd_': 'display_action',
			'm_': 'scroll_action',
			'i_': 'item_action',
		}

		def make_key_mappings(scroll):
			if scroll:
				for k in self.scroll_keys['vi']:
					assert k not in self.key_mappings, f'{k!r} is in key_mappings'
				self.key_mappings.update(self.scroll_keys['vi'])
				self.key_mappings.update(self.scroll_keys[sys.platform])
			return self.key_mappings

		scroll = self.scroll = self.cfg.scroll

		key_mappings = make_key_mappings(scroll)
		action_classes = {k: getattr(self, action_map[v[:2]])() for k, v in key_mappings.items()}
		action_methods = {k: getattr(v, key_mappings[k]) for k, v in action_classes.items()}
		prompt = self.prompt_fs.strip().format(
			s='\nScrolling: k=up, j=down, b=pgup, f=pgdown, g=top, G=bottom' if scroll else '')

		self.prompt_width = max(len(l) for l in prompt.split('\n'))
		self.prompt_height = len(prompt.split('\n'))
		self.oneshot_msg = ''
		prompt += '\b'

		clear_screen = '\n\n' if self.cfg.no_blank else CUR_HOME + ('' if scroll else ERASE_ALL)

		from ..term import get_term, get_char, get_char_raw

		if scroll:
			self.term = get_term()
			self.term.register_cleanup()
			self.term.set('noecho')
			get_char = get_char_raw
			msg_r(CUR_HOME + ERASE_ALL)

		while True:

			if self.oneshot_msg and scroll:
				msg_r(self.blank_prompt + self.oneshot_msg + ' ') # oneshot_msg must be a one-liner
				await asyncio.sleep(2)
				msg_r('\r' + ''.ljust(self.term_width))

			reply = get_char(
				clear_screen
				+ await self.format('squeezed', interactive=True, scroll=scroll)
				+ '\n\n'
				+ (self.oneshot_msg + '\n\n' if self.oneshot_msg and not scroll else '')
				+ prompt,
				immed_chars = key_mappings)

			self.oneshot_msg = ''

			if reply in key_mappings:
				ret = action_classes[reply].run(self, action_methods[reply])
				if type(ret).__name__ == 'coroutine':
					await ret
			elif reply == 'q':
				msg('')
				if self.scroll:
					self.term.set('echo')
				return
			else:
				if not scroll:
					msg_r('\ninvalid keypress ')
				await asyncio.sleep(0.3)

	@property
	def blank_prompt(self):
		return CUR_HOME + CUR_DOWN(self.term_height - self.prompt_height) + ERASE_ALL

	def keypress_confirm(self, *args, **kwargs):
		from ..ui import keypress_confirm
		if keypress_confirm(self.cfg, *args, no_nl=self.scroll, **kwargs):
			return True
		else:
			if self.scroll:
				msg_r('\r'+''.ljust(self.term_width)+'\r'+yellow('Canceling! '))
			return False

	class action:

		@enable_echo
		async def run(self, parent, action_method):
			return await action_method(parent)

		async def a_print_detail(self, parent):
			return await self._print(parent, output_type='detail')

		async def a_print_squeezed(self, parent):
			return await self._print(parent, output_type='squeezed')

		async def _print(self, parent, output_type):

			if not parent.disp_data:
				return None

			outfile = '{a}{b}-{c}{d}[{e}].out'.format(
				a = parent.dump_fn_pfx,
				b = f'-{output_type}' if len(parent.print_output_types) > 1 else '',
				c = parent.proto.dcoin,
				d = ('' if parent.proto.network == 'mainnet' else '-'+parent.proto.network.upper()),
				e = ','.join(parent.sort_info(include_group=False)).replace(' ', ''))

			print_hdr = getattr(parent.display_type, output_type).print_header.format(parent.cols)

			msg_r(parent.blank_prompt if parent.scroll else '\n')

			from ..fileutil import write_data_to_file
			from ..exception import UserNonConfirmation
			try:
				write_data_to_file(
					cfg     = parent.cfg,
					outfile = outfile,
					data    = print_hdr + await parent.format(
						display_type    = output_type,
						line_processing = 'print',
						color           = False),
					desc    = f'{parent.desc} listing')
			except UserNonConfirmation:
				parent.oneshot_msg = yellow(f'File {outfile!r} not overwritten by user request')
			else:
				parent.oneshot_msg = green(f'Data written to {outfile!r}')

		async def a_view(self, parent):
			from ..ui import do_pager
			parent.use_cached = True
			msg_r(CUR_HOME)
			do_pager(await parent.format('squeezed', color=True))

		async def a_view_detail(self, parent):
			from ..ui import do_pager
			msg_r(CUR_HOME)
			do_pager(await parent.format('detail', color=True))

	class item_action:

		@enable_echo
		async def run(self, parent, action_method):

			if not parent.disp_data:
				return

			from ..ui import line_input
			while True:
				msg_r(parent.blank_prompt if parent.scroll else '\n')
				ret = line_input(
					parent.cfg,
					f'Enter {parent.item_desc} number (or ENTER to return to main menu): ')
				if ret == '':
					if parent.scroll:
						msg_r(CUR_UP(1) + '\r' + ''.ljust(parent.term_width))
					return
				idx = get_obj(MMGenIdx, n=ret, silent=True)
				if not idx or idx < 1 or idx > len(parent.disp_data):
					msg_r(
						'Choice must be a single number between 1 and {n}{s}'.format(
							n = len(parent.disp_data),
							s = ' ' if parent.scroll else ''))
					if parent.scroll:
						await asyncio.sleep(1.5)
						msg_r(CUR_UP(1) + '\r' + ERASE_ALL)
				else:
					# action return values:
					#  True:   action successfully performed
					#  None:   action aborted by user or no action performed
					#  False:  an error occurred
					#  'redo': user will be re-prompted for item number
					ret = await action_method(parent, idx)
					if ret != 'redo':
						break
					await asyncio.sleep(0.5)

			if parent.scroll and ret is False:
				# error messages could leave screen in messy state, so do complete redraw:
				msg_r(
					CUR_HOME + ERASE_ALL +
					await parent.format(display_type='squeezed', interactive=True, scroll=True))

		async def i_balance_refresh(self, parent, idx):
			if not parent.keypress_confirm(
					f'Refreshing tracking wallet {parent.item_desc} #{idx}. OK?'):
				return 'redo'
			msg_r('Refreshing balance...')
			res = await parent.twctl.get_balance(parent.disp_data[idx-1].addr, force_rpc=True)
			if res is None:
				parent.oneshot_msg = red(
					f'Unable to refresh {parent.proto.dcoin} balance for {parent.item_desc} #{idx}')
				return False
			else:
				await parent.get_data()
				parent.oneshot_msg = yellow(
					f'{parent.proto.dcoin} balance for {parent.item_desc} #{idx} refreshed')
				if res == 0:
					return False # zeroing balance may mess up display

		async def i_addr_delete(self, parent, idx):
			if not parent.keypress_confirm(
					'Removing {} {} from tracking wallet. OK?'.format(
						parent.item_desc, red(f'#{idx}'))):
				return 'redo'
			if await parent.twctl.remove_address(parent.disp_data[idx-1].addr):
				await parent.get_data()
				parent.oneshot_msg = yellow(f'{capfirst(parent.item_desc)} #{idx} removed')
				return True
			else:
				await asyncio.sleep(3)
				parent.oneshot_msg = red('Address could not be removed')
				return False

		async def i_comment_add(self, parent, idx):

			async def do_comment_add(comment):

				if await parent.twctl.set_comment(
						addrspec     = None,
						comment      = comment,
						trusted_pair = (entry.twmmid, entry.addr),
						silent       = parent.scroll):
					entry.comment = comment
					edited = cur_comment and comment
					parent.oneshot_msg = (green if comment else yellow)('Label {a} {b}{c}'.format(
						a = 'for' if edited else 'added to' if comment else 'removed from',
						b = desc,
						c = ' edited' if edited else ''))
					return True
				else:
					await asyncio.sleep(3)
					parent.oneshot_msg = red('Label for {desc} could not be {action}'.format(
						desc = desc,
						action = 'edited' if cur_comment and comment else 'added' if comment else 'removed'
					))
					return False

			entry = parent.disp_data[idx-1]
			desc = f'{parent.item_desc} #{idx}'
			cur_comment = parent.disp_data[idx-1].comment
			msg('Current label: {}'.format(cur_comment.hl() if cur_comment else '(none)'))

			from ..ui import line_input
			res = line_input(
				parent.cfg,
				'Enter label text for {} {}: '.format(parent.item_desc, red(f'#{idx}')),
				insert_txt = cur_comment)

			if res == cur_comment:
				parent.oneshot_msg = yellow(f'Label for {desc} unchanged')
				return None
			elif res == '':
				if not parent.keypress_confirm(f'Removing label for {desc}. OK?'):
					return 'redo'

			return await do_comment_add(res)

	class scroll_action:

		def run(self, parent, action_method):
			self.use_cached = True
			return action_method(parent)

		def m_cursor_up(self, parent):
			parent.pos -= min(parent.pos - 0, 1)

		def m_cursor_down(self, parent):
			parent.pos += min(parent.max_pos - parent.pos, 1)

		def m_pg_up(self, parent):
			parent.pos -= min(parent.scrollable_height, parent.pos - 0)

		def m_pg_down(self, parent):
			parent.pos += min(parent.scrollable_height, parent.max_pos - parent.pos)

		def m_top(self, parent):
			parent.pos = 0

		def m_bot(self, parent):
			parent.pos = parent.max_pos

	class sort_action:

		def run(self, parent, action_method):
			return action_method(parent)

		def s_addr(self, parent):
			parent.do_sort('addr')

		def s_age(self, parent):
			parent.do_sort('age')

		def s_amt(self, parent):
			parent.do_sort('amt')

		def s_txid(self, parent):
			parent.do_sort('txid')

		def s_twmmid(self, parent):
			parent.do_sort('twmmid')

		def s_reverse(self, parent):
			parent.data.reverse()
			parent.reverse = not parent.reverse

	class display_action:

		def run(self, parent, action_method):
			return action_method(parent)

		def d_days(self, parent):
			af = parent.age_fmts
			parent.age_fmt = af[(af.index(parent.age_fmt) + 1) % len(af)]
			if parent.update_widths_on_age_toggle: # TODO
				pass

		def d_redraw(self, parent):
			msg_r(CUR_HOME + ERASE_ALL)

		def d_addr_view_pref(self, parent):
			parent.addr_view_pref = (parent.addr_view_pref + 1) % len(parent.bch_addr_fmts)



# The reference Ed25519 software is in the public domain.
#     Source: https://ed25519.cr.yp.to/python/ed25519.py
#     Date accessed: 2 Nov. 2016

b = 256
q = 2**255 - 19
l = 2**252 + 27742317777372353535851937790883648493

def expmod(b, e, m):
	if e == 0:
		return 1
	t = expmod(b, e//2, m)**2 % m
	if e & 1:
		t = (t*b) % m
	return t

def inv(x):
	return expmod(x, q-2, q)

d = -121665 * inv(121666)
I = expmod(2, (q-1)//4, q)

def xrecover(y):
	xx = (y*y-1) * inv(d*y*y+1)
	x = expmod(xx, (q+3)//8, q)
	if (x*x - xx) % q != 0:
		x = (x*I) % q
	if x % 2 != 0:
		x = q-x
	return x

By = 4 * inv(5)
Bx = xrecover(By)
B = [Bx%q, By%q]

def edwards(P, Q):
	x1 = P[0]
	y1 = P[1]
	x2 = Q[0]
	y2 = Q[1]
	x3 = (x1*y2+x2*y1) * inv(1+d*x1*x2*y1*y2)
	y3 = (y1*y2+x1*x2) * inv(1-d*x1*x2*y1*y2)
	return [x3%q, y3%q]

def scalarmult(P, e):
	if e == 0:
		return [0, 1]
	Q = scalarmult(P, e//2)
	Q = edwards(Q, Q)
	if e & 1:
		Q = edwards(Q, P)
	return Q

def encodepoint(P):
	x = P[0]
	y = P[1]
	bits = [(y >> i) & 1 for i in range(b-1)] + [x & 1]
	return bytes([sum(bits[i * 8 + j] << j for j in range(8)) for i in range(b//8)])
